import re
from functools import cached_property

from hotsos.core.issues import IssuesManager, OpenvSwitchWarning
from hotsos.core.plugins.openvswitch import OpenvSwitchBase
from hotsos.core.plugins.openvswitch.common import (
    OpenvSwitchEventHandlerBase,
    OpenvSwitchEventCallbackBase,
)
from hotsos.core.utils import sorted_dict


class OVSEventCallbackVSwitchd(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS vswitchd events """
    event_group = 'ovs'
    event_names = ['bridge-no-such-device', 'netdev-linux-no-such-device']

    def __call__(self, event):
        ret = self.categorise_events(
            event,
            options=self.EventProcessingOptions(max_results_per_date=5)
        )
        if ret:
            return {event.name: ret}, 'ovs-vswitchd'

        return None


class OVSEventCallbackLogs(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS log events """
    event_group = 'ovs'
    event_names = ['ovsdb-server', 'ovs-vswitchd',
                   'receive-tunnel-port-not-found',
                   'rx-packet-on-unassociated-datapath-port',
                   'dpif-netlink-lost-packet-on-handler',
                   'assertion-failures',
                   'unreasonably-long-poll-interval']

    def __call__(self, event):
        options = self.EventProcessingOptions(squash_if_none_keys=True)

        if event.name in ['ovs-vswitchd', 'ovsdb-server']:
            options.key_by_date = False

        ret = self.categorise_events(event, options=options)
        if ret:
            return {event.name: ret}, event.section_name

        return None


class OVSEventCallbackDALR(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS dalr events """
    event_group = 'ovs'
    event_names = ['deferred-action-limit-reached']

    def __call__(self, event):
        # pylint: disable=duplicate-code
        results = [{'date': f"{r.get(1)} {r.get(2)}",
                    'time': r.get(3),
                    'key': r.get(4)} for r in event.results]
        ret = self.categorise_events(
            event,
            results=results,
            options=self.EventProcessingOptions(key_by_date=False)
        )
        if ret:
            return {event.name: ret}, event.section_name

        return None


class OVSEventCallbackPortStats(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS portstat events """
    event_group = 'ovs'
    event_names = ['port-stats']

    @cached_property
    def ovs(self):
        return OpenvSwitchBase()

    def _get_port_stats(self, section, sequence_def):
        """
        Extract stats info from port section.

        @param section: searchkit.SearchResult object that is part of a
                        sequence section and is expected to have a tag matching
                        the start or body of a section.
        @param sequence_def: the searchkit.SequenceSearchDef object used to
                             define the search used for this event. This object
                             contains a unique id that is used to identify
                             results.
        @return: if successful returns a tuple of (portname, stats) where stats
                 is a dictionary of key network statistics info about the port.
        """
        port = None
        stats = {}
        for result in section:
            if result.tag == sequence_def.start_tag:
                port = result.get(1)
                continue

            if result.tag != sequence_def.body_tag:
                continue

            key = result.get(1)  # this is rx/tx
            packets = int(result.get(2))
            errors = int(result.get(3))
            dropped = int(result.get(4))

            log_stats = errors or dropped
            if packets:
                dropped_pcent = int((100 / packets) * dropped)
                errors_pcent = int((100 / packets) * errors)
                if max([dropped_pcent, errors_pcent]) <= 1:
                    log_stats = False

            if log_stats:
                stats[key] = {"packets": packets}
                if errors:
                    stats[key]["errors"] = errors
                if dropped:
                    stats[key]["dropped"] = dropped

        if not (port and stats):
            return None, None

        # Ports to ignore - see docstring for info
        if (port in [b.name for b in self.ovs.bridges] or
                re.compile(r"^(q|s)g-\S{11}$").match(port)):
            return None, None

        return port, stats

    def __call__(self, event):
        """
        Report on interfaces that are showing packet drops or errors.

        Sometimes it is normal for an interface to have packet drops and if
        we think that is the case we ignore but otherwise we raise an issue
        to alert.

        Interfaces we currently ignore:

        OVS bridges.

        Neutron HA gateway ports:
        In Openstack for example when using Neutron HA routers, vrrp peers
        that are in BACKUP state may still receive packets on their external
        interface but these will be dropped since they have no where to go. In
        this case it is possible to have 100% packet drops on the interface
        if that VR has never been a vrrp MASTER. For this scenario we filter
        interfaces whose name matches e.g. qg-3ca935f4-07.

        @param event: EventCheckResult object
        """
        port_stats = {}
        global_stats = {'all_dropped': [],  # all interfaces with 100% dropped
                        'all_errors': []}  # all interfaces with 100% errors
        for section in event.results:
            port, _stats = self._get_port_stats(section, event.sequence_def)
            if not _stats:
                continue

            for key in ['dropped', 'errors']:
                for _, s in _stats.items():
                    if s.get(key) and not s['packets']:
                        global_stats['all_' + key].append(port)

            port_stats[port] = _stats

        if not port_stats:
            return None

        ports = global_stats['all_' + 'dropped']
        if ports:
            msg = (f"found {len(ports)} ovs interfaces with 100% "
                   "dropped packets.")
            IssuesManager().add(OpenvSwitchWarning(msg))

        ports = global_stats['all_' + 'errors']
        if ports:
            msg = (f"found {len(ports)} ovs interfaces with 100% "
                   "packet errors.")
            IssuesManager().add(OpenvSwitchWarning(msg))

        stats_sorted = {}
        for k in sorted(port_stats):
            stats_sorted[k] = port_stats[k]

        output_key = f"{event.section_name}-port-stats"
        return stats_sorted, output_key


class OVSEventCallbackBFD(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS BFD events """
    event_group = 'ovs'
    event_names = ['bfd-state-changes']

    def __call__(self, event):
        state_changes = {}
        for r in event.results:
            date = r.get(1)
            port = r.get(2)
            statechange = r.get(3)

            if date not in state_changes:
                state_changes[date] = {}

            if port not in state_changes[date]:
                state_changes[date][port] = []

            state_changes[date][port].append(statechange)

        stats = {}
        if state_changes:
            stats['all-ports-day-avg'] = {}
            stats['per-port-day-total'] = {}
            for date, ports in state_changes.items():
                port_sum = 0
                if date not in stats['per-port-day-total']:
                    stats['per-port-day-total'][date] = {}

                port_totals = {}
                for port, state_changes in ports.items():
                    num_changes = len(state_changes)
                    port_totals[port] = num_changes
                    port_sum += num_changes

                # sort by value
                port_totals = sorted_dict(port_totals, key=lambda e: e[1],
                                          reverse=True)
                stats['per-port-day-total'][date] = port_totals
                day_avg = int(port_sum / len(ports))
                stats['all-ports-day-avg'][date] = day_avg

        stats['all-ports-day-avg'] = sorted_dict(stats['all-ports-day-avg'])
        if stats:
            return {'bfd': {'state-change-stats': stats}}, 'ovs-vswitchd'

        return None


class OVSEventCallbackInvoluntaryContextSwitches(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS involuntary context switch events """
    event_group = 'ovs'
    event_names = ['involuntary-context-switches']

    def __call__(self, event):
        aggregated = {}
        for r in event.results:
            key = r.get(1)
            hour = r.get(2)
            count = int(r.get(3))
            if key not in aggregated:
                aggregated[key] = {}

            if hour in aggregated[key]:
                aggregated[key][hour] += count
            else:
                aggregated[key][hour] = count

        for key, value in aggregated.items():
            aggregated[key] = sorted_dict(value)

        return {event.name: aggregated}, event.section_name


class OVSEventChecks(OpenvSwitchEventHandlerBase):
    """ Events handler for OVS events """
    event_group = 'ovs'
    summary_part_index = 1

    @property
    def summary_subkey(self):
        return 'ovs-checks'


class OVNEventCallbackLogs(OpenvSwitchEventCallbackBase):
    """ Events callback for OVS log events """
    event_group = 'ovn'
    event_names = ['ovsdb-server-nb', 'ovsdb-server-sb', 'ovn-northd',
                   'ovn-controller', 'unreasonably-long-poll-interval',
                   'inactivity-probe', 'bridge-not-found-for-port',
                   'leadership-transfers', 'compactions',
                   'leadership-acquired']

    def __call__(self, event):
        options = self.EventProcessingOptions(squash_if_none_keys=True)

        if event.name in ['ovsdb-server-nb', 'ovsdb-server-sb', 'ovn-northd',
                          'ovn-controller']:
            options.key_by_date = False

        ret = self.categorise_events(event, options=options)
        if ret:
            return {event.name: ret}, event.section_name

        return None


class OVNEventCallbackContextSwitches(OpenvSwitchEventCallbackBase):
    """ Events callback for OVN involuntary context switch events """
    event_group = 'ovn'
    event_names = ['involuntary-context-switches']

    def __call__(self, event):
        aggregated = {}
        for r in event.results:
            key = r.get(1)
            hour = r.get(2)
            count = int(r.get(3))
            if key not in aggregated:
                aggregated[key] = {}

            if hour in aggregated[key]:
                aggregated[key][hour] += count
            else:
                aggregated[key][hour] = count

        for key, value in aggregated.items():
            aggregated[key] = sorted_dict(value)

        return {event.name: sorted_dict(aggregated)}, event.section_name


class OVNEventChecks(OpenvSwitchEventHandlerBase):
    """ Events handler for OVN events """
    event_group = 'ovn'
    summary_part_index = 2

    @property
    def summary_subkey(self):
        return 'ovn-checks'



import os
import re

from hotsos.core.config import HotSOSConfig
from hotsos.core import host_helpers
from hotsos.core.plugins.system.system import SystemBase


class SYSFSBase():
    """ Base class for sysfs interface implementations. """
    @staticmethod
    def get(relpath):
        """
        Read a sysfs entry and return its value.

        @param relpath: path relative to <data_root>sys
        """
        path = os.path.join(HotSOSConfig.data_root, 'sys', relpath)
        if not os.path.exists(path):
            return None

        with open(path, encoding='utf-8') as fd:
            return fd.read().strip()


class CPU(SYSFSBase):
    """ Helper to get CPU information. """
    @property
    def model(self):
        out = host_helpers.CLIHelper().lscpu()
        if out:
            for line in out:
                if not line.startswith("Model name:"):
                    continue

                return re.search(r'Model name:\s+(.+)', line).group(1)

        return None

    @property
    def vendor(self):
        out = host_helpers.CLIHelper().lscpu()
        if out:
            for line in out:
                if not line.startswith("Vendor ID:"):
                    continue

                return re.search(r'Vendor ID:\s+(.+)', line).group(1).lower()

        return None

    @property
    def isolated(self):
        """ This means that isolcpus is configured. """
        return self.get('devices/system/cpu/isolated')

    @property
    def smt(self):
        smt = self.get('devices/system/cpu/smt/active')
        return smt == '1'

    def cpufreq_scaling_governor(self, cpu_id):
        return self.get(f'devices/system/cpu/cpu{cpu_id}/cpufreq/'
                        'scaling_governor')

    @property
    def cpufreq_scaling_governor_all(self):
        governors = set()
        for cpu_id in range(SystemBase().num_cpus):
            cpu_governor = self.cpufreq_scaling_governor(cpu_id)
            if cpu_governor:
                governors.add(cpu_governor)
            else:
                governors.add('unknown')

        return ','.join(list(governors))



import abc
from functools import cached_property
from dataclasses import dataclass

from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.utils import sorted_dict
from hotsos.core.ycheck.engine import (
    YDefsLoader,
    YHandlerBase,
    YDefsSection,
)
from hotsos.core.ycheck.common import GlobalSearcherPreloaderBase
from hotsos.core.exceptions import (
    NotEnoughParametersError,
    AlreadyLoadedError,
    NoCallbacksRegisteredError,
)

CALLBACKS = {}


class EventCallbackNameConflict(Exception):
    """
    When more than one callback are registered with the same name this
    exception is raised.
    """


class EventCallbackNotFound(Exception):
    """
    When an event is raised but the associated callback is not found, this
    exception is raised.
    """


class EventCallbackAutoRegister(type):
    """
    Event callbacks are registered by implementing this metaclass so that they
    are automatically added to the list of callbacks on which we do a lookup to
    find a match when an event is raised.
    """
    def __init__(cls, _name, _mro, members):
        event_group = members.get('event_group')
        if event_group is None:
            return

        for event in members['event_names']:
            event = f'{event_group}.{event}'
            if event in CALLBACKS:
                msg = f"event callback already registered: {event}"
                raise EventCallbackNameConflict(msg)

            CALLBACKS[event] = cls


@dataclass(frozen=True)
class EventCheckResult:
    """ This is passed to an event check callback when matches are found.

        @param name: event label/name from yaml
        @param section_name: section name from yaml
        @param results: searchkit.SearchResultsCollection
        @param search_tag: unique tag used to identify the results
        @param searcher: global FileSearcher object
        @param sequence_def: if set the search results are from a
                            searchkit.SequenceSearchDef and are therefore
                            grouped as sections of results rather than a single
                            set of results.
    """
    name: str
    section_name: str
    results: object
    search_tag: str
    searcher: object
    sequence_def: object = None


class EventProcessingUtils():
    """
    A set of helper methods to help with the processing of event results.
    """

    @dataclass()
    class EventProcessingOptions:
        """Common options for EventProcessingUtils functions.

        @param key_by_date: by default results are categorised by date but in
                            situations where a small number of event types
                            (keys) are spread across many dates/times it might
                            make sense to categorise by event type.
        @param include_time: If true events will be categorised by time beneath
                             date to provide an extra level of granularity.
        @param squash_if_none_keys: If true and any key is found to be None
                                    (perhaps because a regex pattern did not
                                    match properly) the results for each date
                                    will be squashed to a number/count of
                                    events.
        @param max_results_per_date: an integer such that if key_by_date is
                                     True this will pick the top N entries
                                     with the highest count.
        @param tally_value_limit_min: If provided, only tally values greater
                                      than this will be included.
        @param sort_tally_by_value: By default we sort tallied results by value
                                    so that we prioritise keys with the highest
                                    tally value. If we want to sort by key we
                                    need to set this to False.
        """
        key_by_date: bool = True
        include_time: bool = False
        squash_if_none_keys: bool = False
        max_results_per_date: int = None
        tally_value_limit_min: int = None
        sort_tally_by_value: bool = True

    @classmethod
    def _get_event_results(cls, event):
        """
        Return dicts with keys 'date', 'key' and 'value' extracted from event
        search results. The result can have between one and three groups to
        match these keys respectively. If only one group (date) this provide a
        tally per date. Two groups provides a tally per key per date etc.
        """
        _fail_count = 0
        # Use these to ensure we only print the message once to avoid spamming
        # the logs.
        msg_flag = True
        for r in event.results:
            # the search expression used much ensure that these are
            # available in order for this to work.
            if len(r) > 2:
                if not set([None, None, None]).intersection([r.get(1),
                                                             r.get(2),
                                                             r.get(3)]):
                    yield {'date': r.get(1), 'time': r.get(2), 'key': r.get(3)}
                else:
                    _fail_count += 1
            elif len(r) == 0:
                msg = (f"result (tag={r.tag}) does not have enough groups "
                       "(min 1) to be categorised - aborting")
                raise NotEnoughParametersError(msg)
            else:
                if msg_flag and len(r) < 2:
                    msg_flag = False
                    # results with just a date will have None for the key
                    log.debug("results with tag=%s have just one group which "
                              "is assumed to be a date so using key=None",
                              r.tag)

                yield {'date': r.get(1), 'key': r.get(2)}

        if _fail_count:
            log.info("%s results ignored from event %s due to insufficient "
                     "fields",
                     _fail_count, event.name)

    @classmethod
    def _get_tally_keys(cls, options, result):
        """
        By default we tally by 'key' unless options.key_by_date is True in
        which case we tally by value i.e. the date.
        """
        if options.key_by_date:
            return (result['date'], result['key'])

        return (result['key'], result['date'])

    @classmethod
    def _get_tally(cls, result, info, options: EventProcessingOptions):
        key, value = cls._get_tally_keys(options, result)
        if (options.tally_value_limit_min is not None and
                int(value) <= options.tally_value_limit_min):
            return

        if key not in info:
            info[key] = {}

        if value is None and options.squash_if_none_keys:
            if info[key] == {}:
                info[key] = 1
            else:
                info[key] += 1
            return

        ts_time = result.get('time')
        if value not in info[key]:
            if ts_time is not None and options.include_time:
                info[key][value] = {}
            else:
                info[key][value] = 0

        if ts_time is not None and options.include_time:
            if ts_time not in info[key][value]:
                info[key][value][ts_time] = 1
            else:
                info[key][value][ts_time] += 1
        else:
            info[key][value] += 1

    @classmethod
    def _sort_results(cls, categorised_results,
                      options: EventProcessingOptions):
        """
        Sort results.

        By default we sort result dates in ascending order then per-date
        results in descending order of value (tally).

        If results are categorised by value (not date) then we sort both in
        descending order.
        """
        log.debug("sorting categorised results")
        if all([options.key_by_date, options.include_time]):
            return {}

        # sort main dict keys
        categorised_results = sorted_dict(categorised_results,
                                          reverse=not options.key_by_date)

        # Sort values if they are dicts.
        for key, value in categorised_results.items():
            # If not using date as key we need to sort the values so that they
            # will have the same order as if they were sorted by date key.
            if not options.key_by_date:
                categorised_results[key] = sorted_dict(value)
                continue

            # sort by value i.e. tally/count
            for key, value in categorised_results.items():
                if not isinstance(value, dict):
                    break

                categorised_results[key] = sorted_dict(value,
                                                       key=lambda e:
                                                           cls._get_sort_key(
                                                                   e,
                                                                   options),
                                                       reverse=True)

        return categorised_results

    @classmethod
    def global_event_tally_time_granularity_override(cls):
        """
        If results are categorised without time included and this returned True
        we checks if HotSOSConfig.event_tally_granularity is set to 'time' and
        if so enforce time included in results.

        Implementations should override this property to return True if they
        want this behaviour.
        """
        return False

    @staticmethod
    def _get_sort_key(item, options):
        """
        Determine whether to use dict item key or value as a sort key.

        Field must be an integer or able to convert to int.

        @param item: a 2-tuple of the form (key, value)
        @param options: EventProcessingOptions
        @return: integer key
        """
        return int(item[1]) if options.sort_tally_by_value else int(item[0])

    @classmethod
    def categorise_events(
        cls,
        event,
        results=None,
        options: EventProcessingOptions = None,
    ):
        """
        Provides a generic way to categorise events. The default is to group
        events by key which is typically some kind of resource id or event
        label, beneath which the events are grouped by date. If a time value is
        also available, events will be grouped by time beneath their date. It
        may sometimes be useful to group events by date at the top level which
        is also supported here.

        The date, time and key values are extracted from each event search
        result and  are expected to be at indexes 1, 2 and 3 respectively. If
        the search result only has two groups it is assumed that the second is
        "key" i.e. no time is given. If the results argument is provided, this
        will be used instead of event.results.

        @param event: EventCheckResult object
        @param results: optional list of results where each item is a dict with
                        keys 'date', 'key' and optionally 'time'.
        @param options: optional EventProcessingOptions.
        @return: dictionary of results
        """
        if options is None:
            options = EventProcessingUtils.EventProcessingOptions()

        if results is None:
            results = cls._get_event_results(event)

        if cls.global_event_tally_time_granularity_override() is True:
            options.include_time = HotSOSConfig.event_tally_granularity == \
                 "time"

        categorised_results = {}
        for r in results:
            cls._get_tally(r, categorised_results, options)

        if not categorised_results:
            return {}

        categorised_results = cls._sort_results(categorised_results, options)
        if not (options.key_by_date and options.max_results_per_date):
            return categorised_results

        shortened = {}
        for date, entries in categorised_results.items():
            if len(entries) <= options.max_results_per_date:
                shortened[date] = entries
                continue

            top_n = dict(sorted(entries.items(),
                                key=lambda e: cls._get_sort_key(e, options),
                                reverse=True)[:options.max_results_per_date])
            shortened[date] = {
                "total": len(entries),
                f"top{options.max_results_per_date}": top_n,
            }

        return shortened


class EventCallbackBase(EventProcessingUtils,
                        metaclass=EventCallbackAutoRegister):
    """ Base class used for all event callbacks. """
    # All implementations must set these and event_group must match that used
    # in the handler.
    event_group = None
    event_names = []

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """


class EventsSearchPreloader(YHandlerBase, GlobalSearcherPreloaderBase):
    """
    Pre-load all searches used in event definitions into a global FileSearcher
    object and execute the search before running any event callbacks.
    """

    def __init__(self, *args, group=None, **kwargs):
        """
        @param group: a group path can be provided to filter a subset of
                      items.
        """
        super().__init__(*args, **kwargs)
        self.event_group = group

    @property
    def filter(self):
        return HotSOSConfig.event_filter

    @cached_property
    def root(self):
        """ Return the root of all scenarios. """
        plugin_defs = YDefsLoader('events',
                                  filter_path=self.event_group).plugin_defs
        return YDefsSection(HotSOSConfig.plugin_name, plugin_defs or {})

    @property
    def events(self):
        """
        Generator for all events.

        Searches are expected to exist within checks so we extract them from
        each event and discover their searches.

        @return: event
        """
        for event in self.root.leaf_sections:
            # NOTE: event is not an override object, it is a PTreeSection
            # that contains the objects i.e. we go looking for the checks etc
            # within this scenario (sub)section.
            if self.skip_filtered(self.filter, event.resolve_path):
                log.info("skipping event %s from pre-search "
                         "(filter=%s)", event.resolve_path, self.filter)
                continue

            yield event

    def _preload_searches(self):
        """
        Find all events that have a search property and load their search into
        the global searcher.
        """
        if self.global_searcher.is_loaded(self.__class__.__name__):
            raise AlreadyLoadedError(
                "event searches have already been loaded into the "
                "global searcher. This operation can only be "
                "performed once.")

        log.debug("started loading event (group_filter=%s) searches into "
                  "searcher (%s)", self.event_group,
                  self.global_searcher.searcher)

        added = 0
        for event in self.events:
            if self.skip_filtered(self.filter, event.resolve_path):
                log.info("skipping event %s from pre-search (filter=%s)",
                         event.resolve_path, self.filter)
                continue

            added += 1
            self._load_item_search(self.global_searcher, event.search,
                                   event.input)

        self.global_searcher.set_loaded(self.__class__.__name__)
        log.debug("identified a total of %s event check searches", added)
        log.debug("finished loading event searches into searcher "
                  "(registry now has %s items)", len(self.global_searcher))

    def run(self):
        log.debug("registered event callbacks:\n%s", '\n'.
                  join(CALLBACKS.keys()))

        # Pre-load all event searches into a global searcher
        self._preload_searches()


class EventHandlerBase(YHandlerBase, EventProcessingUtils):
    """
    Root name used to identify a group of event definitions. Once all the
    yaml definitions are loaded this defines the level below which events
    for this checker are expected to be found.
    """
    event_group = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # It is assumed that the global searcher already exists, is loaded with
        # searches and they have been executed. Unit tests however, should be
        # resetting the registry prior to each run and we will therefore need
        # to load searches each time which is why we do this here. This is
        # therefore not intended to be used outside of a test scenario.
        label = EventsSearchPreloader.__name__
        if not self.global_searcher.is_loaded(label):
            log.info("global searcher catalog is empty so launching "
                     "pre-load of event searches for group '%s'",
                     self.event_group)
            # NOTE: this is not re-entrant safe and is only ever expected
            #       to be done from a unit test.
            EventsSearchPreloader(self.global_searcher,
                                  group=self.event_group).run()

        self._event_results = None

    @property
    def filter(self):
        return HotSOSConfig.event_filter

    @staticmethod
    def meets_requirements(item):
        """
        If an item or group has a requirements property it must return True
        in order to be executed.
        """
        if HotSOSConfig.force_mode:
            return True

        if item.requires and not item.requires.result:
            log.debug("item '%s' pre-requisites not met - "
                      "skipping", item.name)
            return False

        return True

    @property
    def searcher(self):
        return self.global_searcher.searcher

    @cached_property
    def events(self):
        """ Load event definitions from yaml. """
        group_defs = YDefsLoader('events',
                                 filter_path=self.event_group).plugin_defs
        group = YDefsSection(HotSOSConfig.plugin_name, group_defs or {})
        log.debug("sections=%s, events=%s",
                  len(list(group.branch_sections)),
                  len(list(group.leaf_sections)))

        _events = {}
        for event in group.leaf_sections:
            if EventsSearchPreloader.skip_filtered(self.filter,
                                                   event.resolve_path):
                log.info("skipping event %s (filter=%s)", event.resolve_path,
                         HotSOSConfig.event_filter)
                continue

            if not self.meets_requirements(event):
                return {}

            log.debug("event: %s", event.name)
            log.debug("input: %s (command=%s)", event.input.paths,
                      event.input.command is not None)

            section_name = event.parent.name
            if section_name not in _events:
                _events[section_name] = {}

            _events[section_name][event.name] = event.search.unique_search_tag

        return _events

    @property
    def final_event_results(self):
        """ Cache of results in case run() is called again. """
        return self._event_results

    @staticmethod
    def _get_event_search_results(global_results, search_tag,
                                  sequence_search_def, passthrough_results):
        if passthrough_results:
            # this is for implementations that have their own means of
            # retrieving results.
            return global_results

        if sequence_search_def is None:
            return global_results.find_by_tag(search_tag)

        search_results = global_results.find_sequence_sections(
                             sequence_search_def)
        if search_results:
            return search_results.values()

        return {}

    def _exec_callback(self, event, section_name, search_results,
                       event_search):
        # We want this to throw an exception if the callback is not
        # defined.
        callback_name = f'{self.event_group}.{event}'
        if callback_name not in CALLBACKS:
            msg = f"no callback found for event {callback_name}"
            raise EventCallbackNotFound(msg)

        callback = CALLBACKS[callback_name]
        event_result = EventCheckResult(
            name=event,
            section_name=section_name,
            results=search_results,
            search_tag=event_search['search_tag'],
            searcher=self.global_searcher.searcher,
            sequence_def=event_search['sequence_search']
        )
        log.debug("executing event %s.%s callback '%s'",
                  event_result.section_name, event, callback_name)
        return callback()(event_result)

    def run(self):
        """
        Process each event and call respective callback functions when results
        where found.
        """

        results = self.global_searcher.results
        if self.final_event_results is not None:
            return self.final_event_results

        if not CALLBACKS:
            raise NoCallbacksRegisteredError(
                "need to register at least one callback for event handler.")

        log.debug("running event(s) for group=%s", self.event_group)
        info = {}
        for section_name, section in self.events.items():
            for event, fullname in section.items():
                event_search = self.global_searcher[fullname]
                seq_def = event_search['sequence_search']
                search_tag = event_search['search_tag']
                passthrough = event_search['passthrough_results']
                search_results = self._get_event_search_results(results,
                                                                search_tag,
                                                                seq_def,
                                                                passthrough)
                if not search_results:
                    log.debug("event %s (tag=%s) did not yield any results",
                              event, event_search['search_tag'])
                    continue

                ret = self._exec_callback(event, section_name, search_results,
                                          event_search)
                if not ret:
                    continue

                # if the return is a tuple it is assumed to be of the form
                # (<output value>, <output key>) where <output key> is used to
                # override the output key for the result which defaults to the
                # event name.
                if isinstance(ret, tuple):
                    out_key = ret[1]
                    ret = ret[0]
                    if not ret:
                        continue
                else:
                    out_key = event

                # Don't clobber results, instead allow them to aggregate.
                if out_key in info:
                    info[out_key].update(ret)
                else:
                    info[out_key] = ret

        if info:
            self._event_results = info
            return info

        return {}



from .common import KernelBase
from .config import SystemdConfig
from .kernlog import CallTraceManager

__all__ = [
    CallTraceManager.__name__,
    KernelBase.__name__,
    SystemdConfig.__name__,
    ]



# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' cinder/exception.py
CINDER_EXCEPTIONS = [
    "ConvertedException",
    "CinderException",
    "VolumeBackendAPIException",
    "VolumeDriverException",
    "BackupDriverException",
    "BackupRestoreCancel",
    "GlanceConnectionFailed",
    "ProgrammingError",
    "NotAuthorized",
    "AdminRequired",
    "PolicyNotAuthorized",
    "ImageNotAuthorized",
    "DriverNotInitialized",
    "Invalid",
    "InvalidSnapshot",
    "InvalidVolumeAttachMode",
    "VolumeAttached",
    "InvalidResults",
    "InvalidInput",
    "InvalidAvailabilityZone",
    "InvalidTypeAvailabilityZones",
    "InvalidVolumeType",
    "InvalidGroupType",
    "InvalidVolume",
    "InvalidContentType",
    "InvalidHost",
    "InvalidParameterValue",
    "InvalidAuthKey",
    "InvalidConfigurationValue",
    "ServiceUnavailable",
    "UnavailableDuringUpgrade",
    "ImageUnacceptable",
    "ImageTooBig",
    "DeviceUnavailable",
    "SnapshotUnavailable",
    "InvalidUUID",
    "InvalidAPIVersionString",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "ValidationError",
    "APIException",
    "APITimeout",
    "RPCTimeout",
    "Duplicate",
    "NotFound",
    "GlanceStoreNotFound",
    "GlanceStoreReadOnly",
    "VolumeNotFound",
    "MessageNotFound",
    "VolumeAttachmentNotFound",
    "VolumeMetadataNotFound",
    "InvalidVolumeMetadata",
    "InvalidVolumeMetadataSize",
    "SnapshotMetadataNotFound",
    "VolumeTypeNotFound",
    "VolumeTypeNotFoundByName",
    "VolumeTypeAccessNotFound",
    "VolumeTypeExtraSpecsNotFound",
    "VolumeTypeInUse",
    "VolumeTypeDeletionError",
    "VolumeTypeDefaultDeletionError",
    "VolumeTypeDefaultMisconfiguredError",
    "VolumeTypeProjectDefaultNotFound",
    "GroupTypeNotFound",
    "GroupTypeNotFoundByName",
    "GroupTypeAccessNotFound",
    "GroupTypeSpecsNotFound",
    "GroupTypeInUse",
    "SnapshotNotFound",
    "ServerNotFound",
    "VolumeSnapshotNotFound",
    "VolumeIsBusy",
    "SnapshotIsBusy",
    "InvalidImageRef",
    "InvalidSignatureImage",
    "ImageSignatureVerificationException",
    "ImageNotFound",
    "ServiceNotFound",
    "ServiceTooOld",
    "WorkerNotFound",
    "WorkerExists",
    "CleanableInUse",
    "ClusterNotFound",
    "ClusterHasHosts",
    "ClusterExists",
    "HostNotFound",
    "SchedulerHostFilterNotFound",
    "SchedulerHostWeigherNotFound",
    "InvalidReservationExpiration",
    "InvalidQuotaValue",
    "QuotaNotFound",
    "QuotaResourceUnknown",
    "ProjectQuotaNotFound",
    "QuotaClassNotFound",
    "QuotaUsageNotFound",
    "OverQuota",
    "FileNotFound",
    "VolumeTypeExists",
    "VolumeTypeAccessExists",
    "VolumeTypeEncryptionExists",
    "VolumeTypeEncryptionNotFound",
    "GroupTypeExists",
    "GroupTypeAccessExists",
    "GroupVolumeTypeMappingExists",
    "MalformedRequestBody",
    "ConfigNotFound",
    "ParameterNotFound",
    "NoValidBackend",
    "QuotaError",
    "VolumeSizeExceedsAvailableQuota",
    "VolumeSizeExceedsLimit",
    "VolumeBackupSizeExceedsAvailableQuota",
    "VolumeLimitExceeded",
    "SnapshotLimitExceeded",
    "UnexpectedOverQuota",
    "BackupLimitExceeded",
    "ImageLimitExceeded",
    "VolumeTypeCreateFailed",
    "VolumeTypeUpdateFailed",
    "GroupTypeCreateFailed",
    "GroupTypeUpdateFailed",
    "GroupLimitExceeded",
    "UnknownCmd",
    "MalformedResponse",
    "FailedCmdWithDump",
    "InvalidConnectorException",
    "GlanceMetadataExists",
    "GlanceMetadataNotFound",
    "ImageDownloadFailed",
    "ExportFailure",
    "RemoveExportException",
    "MetadataUpdateFailure",
    "MetadataCopyFailure",
    "InvalidMetadataType",
    "ImageCopyFailure",
    "BackupInvalidCephArgs",
    "BackupOperationError",
    "BackupMetadataUnsupportedVersion",
    "BackupMetadataNotFound",
    "VolumeMetadataBackupExists",
    "BackupRBDOperationFailed",
    "EncryptedBackupOperationFailed",
    "BackupNotFound",
    "InvalidBackup",
    "SwiftConnectionFailed",
    "TransferNotFound",
    "VolumeMigrationFailed",
    "SSHInjectionThreat",
    "QoSSpecsExists",
    "QoSSpecsCreateFailed",
    "QoSSpecsUpdateFailed",
    "QoSSpecsNotFound",
    "QoSSpecsAssociateFailed",
    "QoSSpecsDisassociateFailed",
    "QoSSpecsKeyNotFound",
    "InvalidQoSSpecs",
    "QoSSpecsInUse",
    "KeyManagerError",
    "ManageExistingInvalidReference",
    "ManageExistingAlreadyManaged",
    "InvalidReplicationTarget",
    "UnableToFailOver",
    "ReplicationError",
    "ReplicationGroupError",
    "ManageExistingVolumeTypeMismatch",
    "ExtendVolumeError",
    "EvaluatorParseException",
    "LockCreationFailed",
    "CappedVersionUnknown",
    "VolumeGroupNotFound",
    "VolumeGroupCreationFailed",
    "VolumeNotDeactivated",
    "VolumeDeviceNotFound",
    "RemoteFSException",
    "RemoteFSConcurrentRequest",
    "RemoteFSNoSharesMounted",
    "RemoteFSNoSuitableShareFound",
    "RemoteFSInvalidBackingFile",
    "NfsException",
    "NfsNoSharesMounted",
    "NfsNoSuitableShareFound",
    "ZoneManagerException",
    "FCZoneDriverException",
    "FCSanLookupServiceException",
    "ZoneManagerNotInitialized",
    "ConsistencyGroupNotFound",
    "InvalidConsistencyGroup",
    "GroupNotFound",
    "InvalidGroup",
    "InvalidGroupStatus",
    "CgSnapshotNotFound",
    "InvalidCgSnapshot",
    "GroupSnapshotNotFound",
    "InvalidGroupSnapshot",
    "InvalidGroupSnapshotStatus",
    "ISCSITargetCreateFailed",
    "ISCSITargetRemoveFailed",
    "ISCSITargetAttachFailed",
    "ISCSITargetDetachFailed",
    "TargetUpdateFailed",
    "ISCSITargetHelperCommandFailed",
    "BadHTTPResponseStatus",
    "BadResetResourceStatus",
    "MetadataAbsent",
    "NotSupportedOperation",
    "AttachmentSpecsNotFound",
    "InvalidName",
    "ServiceUserTokenNoAuth",
    "RekeyNotSupported",
    "ImageCompressionNotAllowed",
    "CinderAcceleratorError",
    "SnapshotLimitReached",
]



import abc
import os

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import (
    CLIExecError,
    CLIHelper,
    CLIHelperFile,
    HostNetworkingHelper,
)
from hotsos.core.search import (
    CommonTimestampMatcher,
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core.log import log

KERNLOG_TS = r'\[\s*\d+\.\d+\]'
KERNLOG_PREFIX = rf'(?:\S+\s+\d+\s+[\d:]+\s+\S+\s+\S+:\s+)?{KERNLOG_TS}'


class CallTraceHeuristicBase():
    """ Defines a common interface for heuristics implementations. """

    @abc.abstractmethod
    def __call__(self):
        """ Ensure callable to get results. """


class CallTraceStateBase():
    """
    A state capture object that allows getting and setting arbitrary state.
    """

    def __init__(self):
        self._info = {}

    def add(self, key, value):
        self._info[key] = value

    def __getattr__(self, key):
        return self._info.get(key)


class TraceTypeBase(abc.ABC):
    """
    This defines a common interface to trace types and should be implemented
    for any trace types we want to capture. Implementations of this object
    are typically registered with a CallTraceManager for processing.
    """

    @property
    @abc.abstractmethod
    def name(self):
        """
        Name used to identify the type of call trace e.g. "oomkiller".
        """

    @property
    @abc.abstractmethod
    def searchdef(self):
        """
        A search definition object (simple or sequence) used to identify this
        call trace.
        """

    @abc.abstractmethod
    def apply(self, results):
        """
        Take results and parse into constituent parts.

        @param results: a SearchResultsCollection object containing the results
        of our search.
        """

    @property
    @abc.abstractmethod
    def heuristics(self):
        """ Return a list of CallTraceHeuristic objects that can be used to
        run checks on any identified call traces. """

    @abc.abstractmethod
    def __len__(self):
        """ Return number of call stacks identified.  """

    @abc.abstractmethod
    def __iter__(self):
        """ Iterate over each call trace found. """


class KernLogSource(CLIHelperFile):
    """
    We want to try different sources for kern logs. We implement CLIHelperFile
    so that we can try the systemd journal first and fall back to other paths
    if not successful.
    """
    def __init__(self):
        super().__init__(catch_exceptions=False)
        self.path = None
        self.attempt = 0
        self.fs_paths = ['var/log/kern.log',
                         'sos_commands/logs/journalctl_--no-pager']
        self.fs_paths = [os.path.join(os.path.join(HotSOSConfig.data_root, f))
                         for f in self.fs_paths]

    def __enter__(self):
        if os.path.exists(os.path.join(HotSOSConfig.data_root,
                                       'var/log/journal')):
            try:
                self.path = self.journalctl(opts='-k')
                log.debug("using journal as source of kernlogs")
            except CLIExecError:
                log.info("Failed to get kernlogs from systemd journal. "
                         "Trying fallback sources.")
        else:
            log.info("systemd journal not available. Trying fallback kernlog "
                     "sources.")

        if not self.path:
            for path in self.fs_paths:
                if os.path.exists(path):
                    if path.endswith('kern.log') and HotSOSConfig.use_all_logs:
                        self.path = f"{path}*"
                    else:
                        self.path = path

                    log.debug("using %s as source of kernlogs", self.path)
                    break
            else:
                paths = ', '.join(self.fs_paths)
                log.warning("no kernlog sources found (tried %s and "
                            "journal)", paths)

        return self


class KernLogBase():
    """ Base class for kernlog analysis implementations. """
    def __init__(self):
        self.hostnet_helper = HostNetworkingHelper()
        self.cli_helper = CLIHelper()

    @staticmethod
    def perform_search(searchdefs):
        try:
            constraint = SearchConstraintSearchSince(
                                         ts_matcher_cls=CommonTimestampMatcher)
        except ValueError as exc:
            log.warning("failed to create global search constraint for "
                        "calltrace checker: %s", exc)
            constraint = None

        searcher = FileSearcher(constraint=constraint)
        with KernLogSource() as kls:
            if kls.path is None:
                log.info("no kernlogs found")
                return None

            for sd in searchdefs:
                searcher.add(sd, kls.path)

            return searcher.run()



from hotsos.core.plugins.pacemaker import PacemakerChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class PacemakerSummary(PacemakerChecks):
    """ Implementation of Pacemaker summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @summary_entry('nodes', get_min_available_entry_index())
    def summary_nodes(self):
        nodes = {}
        if self.online_nodes:
            nodes["online"] = self.online_nodes
        if self.offline_nodes:
            nodes["offline"] = self.offline_nodes

        return nodes or None



import os
from dataclasses import dataclass, field
from functools import cached_property

from hotsos.core.host_helpers import (
    InstallInfoBase,
    PebbleHelper,
    SystemdHelper,
)
from hotsos.core.plugins.juju.resources import JujuBase
from hotsos.core import plugintools

SVC_VALID_SUFFIX = r'[0-9a-zA-Z-_]*'
JUJU_SVC_EXPRS = [rf'mongod{SVC_VALID_SUFFIX}',
                  rf'jujud{SVC_VALID_SUFFIX}',
                  # catch juju-db but filter out processes with juju-db in
                  # their args list.
                  rf'(?:^|[^\s])juju-db{SVC_VALID_SUFFIX}']


@dataclass
class JujuInstallInfo(InstallInfoBase):
    """ Juju installation information. """
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(service_exprs=JUJU_SVC_EXPRS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=JUJU_SVC_EXPRS))


class JujuChecks(plugintools.PluginPartBase, JujuBase):
    """ Juju checks. """
    plugin_name = 'juju'
    plugin_root_index = 12

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        JujuInstallInfo().mixin(self)

    @cached_property
    def systemd_processes(self):
        """
        Return a list of running processes related to the Juju service. This
        is needed for Juju scenarios.
        """
        return self.systemd.processes

    @property
    def version(self):
        if self.machine:
            return self.machine.version

        return "unknown"

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        return os.path.exists(cls.get_juju_lib_path())



import builtins
import importlib
import inspect
from collections import UserDict

from propertree.propertree2 import (
    PTreeOverrideBase,
    PTreeMappedOverrideBase,
    PTreeSection,
)
from hotsos.core.log import log
from hotsos.core.host_helpers.config import ConfigBase
from hotsos.core.exceptions import (
    NotEnoughParametersError,
    MissingRequiredParameterError,
    UnexpectedParameterError
)


class ImportPathIsNotAClass(Exception):
    """ Exception used when an import is not a class. """
    def __init__(self, msg):
        self.msg = msg
        super().__init__()


class YDefsSection(PTreeSection):
    """
    Implementation of PTreeSection that ensures a context is always provided.
    This is used by all YAML handler types to load their defintitions e.g. of
    events or scenarios.
    """
    def __init__(self, name, content, *args, context=None, **kwargs):
        """
        @param name: name of defs group
        @param content: defs tree of type dict
        @param context: optional context object. This gets passed to all
                        resolved properties and can be used to share
                        information amongst them. If not provided a new empty
                        one is created.
        """
        super().__init__(name, content, *args,
                         context=context or YDefsContext(), **kwargs)


class YDefsContext(UserDict):
    """
    Provides a way to get/set arbitrary information used as context to a yaml
    defs run. This object is typically passed to a YDefsSection and accessible
    from all override properties as attribute "context".
    """
    def __init__(self, initial_state=None):
        """
        @param initial_state: optional dict to use as initial state.
        """
        super().__init__(initial_state or {})

    def __setitem__(self, key, item):
        # don't print values as they can be very large and e.g. search
        # results will be expanded re-duplicated.
        log.debug("%s setting %s with value of type '%s'",
                  self.__class__.__name__, key, type(item))
        super().__setitem__(key, item)

    def __getattr__(self, key):
        return self.data.get(key)


class PropertyCacheRefResolver():
    """
    This class is used to resolve string references to property cache entries.
    """
    def __init__(self, refstr, pcrr_vars=None, checks=None):
        log.debug("%s: resolving '%s'", self.__class__.__name__, refstr)
        self.refstr = refstr
        if not self.is_valid_cache_ref(refstr):
            msg = (f"{refstr} is not a valid property cache reference or "
                   "variable")
            raise NameError(msg)

        self.vars = pcrr_vars
        self.checks = checks
        if self.reftype == 'checks' and checks is None:
            msg = (f"{refstr} is a checks cache reference but checks dict not "
                   "provided")
            raise NotEnoughParametersError(msg)

    @property
    def reftype(self):
        """
        These depict the type of property or propertycollection that can be
        referenced.

        Supported formats:
            $varname[:func]
            @checks.<check_name>.<property_name>.<property_cache_key>[:func]
        """
        if self.refstr.startswith('$'):
            return "variable"
        if self.refstr.startswith('@checks.'):
            # This is an implementation of YPropertyChecks
            return "checks"

        raise NameError(f"unknown ref type {self.refstr}")

    @property
    def _ref_body(self):
        """
        Strip the prefix from the reference string.
        """
        if self.reftype == 'variable':
            prefix = "$"
        elif self.reftype == 'checks':
            prefix = f"@checks.{self.check_name}."

        return self.refstr.partition(prefix)[2]

    @classmethod
    def is_valid_cache_ref(cls, refstr):
        """
        Returns True if refstr is a valid property cache reference.

        The criteria for a valid reference is that it must be a string whose
        first character is @.
        """
        if not isinstance(refstr, str):
            return False

        if not (refstr.startswith('@') or refstr.startswith('$')):
            return False

        return True

    @property
    def check_name(self):
        if self.reftype != 'checks':
            raise MissingRequiredParameterError(
                "ref does not have type 'checks'")

        return self.refstr.partition('@checks.')[2].partition('.')[0]

    @property
    def property_name(self):
        if self.reftype != 'checks':
            return None

        return self._ref_body.partition('.')[0]

    @property
    def property_cache_key(self):
        """ Key for PropertyCache. """
        if self.reftype == 'checks':
            _key = self._ref_body.partition('.')[2]
        else:
            raise UnexpectedParameterError(
                f"not a check reftype: {self.reftype}")

        # strip func if exists
        return _key.partition(':')[0]

    @property
    def property_cache_value_renderer_function(self):
        """
        This is an optional function name that can be provided as the last
        item in the reference string seperated by a colon.
        """
        if self.reftype == 'checks':
            _key = self._ref_body.partition('.')[2]
        else:
            _key = self._ref_body

        return _key.partition(':')[2]

    @staticmethod
    def apply_renderer(value, func):
        """
        A cache reference or import path can be suffixed with :<function>
        where <function> can be one of two things; either a builtins function
        or one of the extras defined here. Obviously not all builtins can be
        used and their use depends on the type() of value.

        @param value: value to apply renderer function to
        @param func: function to apply to value
        """
        extras = {'comma_join': {
                    'requirements': [lambda value: isinstance(value, list),
                                     lambda value: isinstance(value, dict)],
                    'action': ', '.join},
                  'unique_comma_join': {
                    'requirements': [lambda value: isinstance(value, list),
                                     lambda value: isinstance(value, dict)],
                    'action': lambda value: ', '.join(sorted(set(value)))},
                  'first': {
                    'requirements': [lambda value: isinstance(value, list)],
                    'action': lambda value: value[0]},
                  'int_ranges': {
                    'requirements': [lambda value: isinstance(value, list)],
                    'action': ConfigBase.squash_int_range}}

        if func in extras:
            if not any(req(value) for req in extras[func]['requirements']):
                log.warning("attempted to apply '%s' to value of "
                            "type %s", func, type(value))
                return value

            return extras[func]['action'](value)

        return getattr(builtins, func)(value)

    def _get_search_result_group(self, cache_key):
        """
        Extract values at the given group index from all search results.
        """
        val = []
        group = int(cache_key.partition('results_group_')[2])
        for result in self.checks[self.check_name].search_results:
            if len(result) >= group:
                val.append(result.get(group))

        return sorted(val)

    def resolve(self):
        """
        Resolve value associated with a check property result or variable.

        Values are then run through an optional renderer function.
        """
        if self.reftype == 'checks':
            cache_key = self.property_cache_key
            # This provides an interface to extract the values of specific
            # search result groups.
            if cache_key.startswith('results_group_'):
                val = self._get_search_result_group(cache_key)
            else:
                check_cache = self.checks[self.check_name].cache
                property_cache = getattr(check_cache, self.property_name)
                val = getattr(property_cache, cache_key)
        else:
            varname = self.refstr.partition("$")[2]
            varname = varname.partition(':')[0]
            val = self.vars.resolve(varname)

        if val is None:
            return None

        func = self.property_cache_value_renderer_function
        if not func:
            # noop
            return val

        return self.apply_renderer(val, func)


class PropertyCache(UserDict):
    """
    A cached used by YAML property objects. This is useful where we want to
    retain state for the lifetime of a property object instance. Note that this
    state does and will not persist a reload of the YAML tree i.e. if the
    property objects are recreated the caches are as well.
    """
    def merge(self, cache):
        if not isinstance(cache, self.__class__):
            log.error("attempt to merge cache failed - provided cache is not "
                      "a %s", type(self.__class__.__name__))
            return

        self.data.update(cache.data)

    @property
    def id(self):
        return id(self)

    def set(self, key, data):
        log.debug("%s: caching key=%s with value=%s", id(self), key, data)
        _current = self.data.get(key)
        if _current and isinstance(_current, dict) and isinstance(data, dict):
            self.data[key].update(data)
        else:
            self.data[key] = data

    def __getattr__(self, key):
        log.debug("%s: fetching key=%s (exists=%s)", self.id, key,
                  key in self.data)
        return self.data.get(key)


class ImportHelper:
    """ Helper class for importing modules/classes/properties from strings. """

    @staticmethod
    def _load_from_import_cache(context, key):
        """ Retrieve from global context if one exists.

        @param context: global context property. This is where we look for the
                        cache which is stored using key 'import_cache'. If an
                        import is cached we use it rather than re-importing.
        @param key: key to retrieve
        """
        if context is None:
            log.info("context not available - cannot load '%s'", key)
            return None

        # we save all imports in a dict called "import_cache" within the
        # global context so that all properties have access.
        c = getattr(context, 'import_cache')
        if not c:
            return None

        return c.get(key)

    @staticmethod
    def _get_mod_class_from_path(path):
        _mod = path.rpartition('.')[0]
        _cls = path.rpartition('.')[2]
        return _mod, _cls

    @staticmethod
    def _get_class_property_from_path(path, no_property=False):
        log.debug("fetching class and property from path (no_property=%s)",
                  no_property)
        # first strip any factory class info and add back to prop at end.
        _path, _, factinput = path.rpartition(':')
        if _path:
            path = _path
        else:
            factinput = ''

        if no_property:
            _cls = path
            _prop = ''
        else:
            _cls = path.rpartition('.')[0]
            _prop = path.rpartition('.')[2]

        # now put it back if exists
        if factinput:
            log.debug("path is factory (input=%s)", factinput)
            _prop += ":" + factinput

        return _cls, _prop

    @staticmethod
    def _add_to_import_cache(context, key, value):
        """ Save in the global context if one exists.

        @param context: global context property. This is where we look for the
                        cache which is stored using key 'import_cache'. If an
                        import is cached we use it rather than re-importing.
        @param key: key to save
        @param value: value to save
        """
        if context is None:
            log.info("context not available - cannot save '%s'", key)
            return

        c = getattr(context, 'import_cache')
        if c:
            c[key] = value
        else:
            c = {key: value}
            setattr(context, 'import_cache', c)

    def get_cls(self, context, import_str):
        """ Import and instantiate Python class.

        @param context: global context property. This is where we look for the
                        cache which is stored using key 'import_cache'. If an
                        import is cached we use it rather than re-importing.
        @param import_str: import path to Python class.
        """
        ret = self._load_from_import_cache(context, import_str)
        if ret:
            log.debug("instantiating class %s (from_cache=True)", import_str)
            return ret

        log.debug("instantiating class %s (from_cache=False)", import_str)
        mod, cls_name = self._get_mod_class_from_path(import_str)
        try:
            _mod = importlib.import_module(mod)
            ret = getattr(_mod, cls_name)
            # NOTE: we don't use isclass() since it doesn't work with
            # mock.MagicMock in tests.
            if inspect.ismodule(ret):
                log.debug("%s is not a class", ret)
                raise ImportPathIsNotAClass(import_str)
        except ImportPathIsNotAClass:
            raise
        except Exception:
            log.exception("failed to import class %s from %s", cls_name, mod)
            raise

        self._add_to_import_cache(context, import_str, ret)
        return ret

    def get_property(self, context, import_str):
        """
        Import and fetch value of a Python property or factory.

        If the path is to a property, the property name is expected to be after
        the last '.' with the preceding name being the parent class.

        A factory path is identified as having a ':' delimiter which denotes
        the input and optional attribute to call on the factory object. In
        this case the field prior to the delim is the path to the factory class
        itself with an optional input.

        @param context: global context property. This is where we look for the
                        cache which is stored using key 'import_cache'. If an
                        import is cached we use it rather than re-importing.
        @param import_str: a path to a Python property or Factory.
        """
        ret = self._load_from_import_cache(context, import_str)
        if ret:
            log.debug("calling property %s (from_cache=True)", import_str)
            return ret

        log.debug("calling property %s (from_cache=False)", import_str)
        _cls, _prop = self._get_class_property_from_path(import_str)
        try:
            try:
                cls = self.get_cls(context, _cls)
            except ImportPathIsNotAClass:
                # support case where factory has no attribute
                log.debug("trying again without property")
                _cls, _prop = self._get_class_property_from_path(
                                                              import_str,
                                                              no_property=True)
                cls = self.get_cls(context, _cls)
        except Exception:
            log.exception("class '%s' import failed", _cls)
            raise

        key = f"{cls}.object"
        cls_inst = self._load_from_import_cache(context, key)
        if not cls_inst:
            try:
                cls_inst = cls(global_searcher=context.global_searcher)
            except TypeError:
                cls_inst = cls()

            self._add_to_import_cache(context, key, cls_inst)

        if ':' in _prop:
            # property is for a factory object
            fattr, _, finput = _prop.partition(':')
            if fattr:
                properties = [finput, fattr]
            else:
                properties = [finput]
        else:
            properties = [_prop]

        _obj = cls_inst
        for _prop in properties:
            try:
                _obj = getattr(_obj, _prop)
            except Exception:
                log.exception("failed to import and call property %s",
                              import_str)

                raise

            if callable(_obj):
                raise TypeError(f"{_obj} is a not a property (looks like it "
                                "is a callable method)")

        self._add_to_import_cache(context, import_str, _obj)
        return _obj

    @staticmethod
    def get_method(import_str):
        """ Import and instantiate Python class then call method.

        @param import_str: import path to Python class with method name after
                           the final '.'.
        """
        log.debug("calling method %s", import_str)
        mod = import_str.rpartition('.')[0]
        method = import_str.rpartition('.')[2]
        class_name = mod.rpartition('.')[2]
        mod = mod.rpartition('.')[0]
        cls = getattr(importlib.import_module(mod), class_name)
        try:
            ret = getattr(cls(), method)()
        except Exception:
            log.exception("failed to import and call method %s", import_str)
            raise

        return ret

    @staticmethod
    def get_attribute(import_str):
        log.debug("fetching attribute %s", import_str)
        mod = import_str.rpartition('.')[0]
        attr = import_str.rpartition('.')[2]
        try:
            ret = getattr(importlib.import_module(mod), attr)
        except Exception as exc:
            log.exception("failed to get module attribute %s", import_str)

            # propertree.PTreeOverrideBase swallows AttributeError so need to
            # convert to something else.
            if isinstance(exc, AttributeError):
                raise ImportError from exc

            raise

        return ret

    def get_import(self, context, import_str):
        """
        First attempt to treat import string as a class property then try
        module attribute.

        @param context: global context property. This is where we look for the
                        cache which is stored using key 'import_cache'. If an
                        import is cached we use it rather than re-importing.
        """
        try:
            return self.get_property(context, import_str)
        except TypeError as exc:
            log.debug("get_property failed, trying get_attribute: %s", exc)

        return self.get_attribute(import_str)


class YPropertyBase(PTreeOverrideBase, ImportHelper):
    """
    Base class used for all YAML property objects. Implements and extends
    PTreeOverrideBase to provide frequently used methods and helpers to
    property implementations.
    """
    def __init__(self, *args, **kwargs):
        self._cache = PropertyCache()
        super().__init__(*args, **kwargs)

    @property
    def cache(self):
        """
        All properties get their own cache object that they can use as they
        wish.
        """
        return self._cache

    def resolve_var(self, name):
        """
        Resolve variable with name to value. This can be used speculatively and
        will return the name as value if it can't be resolved.
        """
        if not name.startswith('$'):
            return name

        if hasattr(self, 'context'):
            if self.context.vars:
                _name = name.partition('$')[2]
                return self.context.vars.resolve(_name)

        log.warning("could not resolve var '%s' - vars not found in "
                    "context", name)

        return name


class YPropertyOverrideBase(YPropertyBase, PTreeOverrideBase):
    """ Base class for all simple/flat property objects. """


class YPropertyMappedOverrideBase(YPropertyBase, PTreeMappedOverrideBase):
    """ Base class for all mapped property objects. """



from datetime import datetime

from hotsos.core.analytics import LogEventStats
from hotsos.core.plugins.openstack.common import (
    OpenstackBase,
    OpenStackChecks,
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from hotsos.core.plugins.openstack.nova import NovaLibvirt
from hotsos.core import utils
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class OpenstackInstanceChecks(OpenstackBase, OpenStackChecks):
    """ Implements Openstack Nova instance checks. """
    summary_part_index = 2

    @summary_entry('vm-info', get_min_available_entry_index() + 1)
    def summary_vm_info(self):
        _info = {}

        instances = self.project_helpers.nova.instances.values()
        if instances:
            _info['running'] = {'count': len(instances),
                                'uuids': [i.uuid for i in instances]}

        novalibvirt = NovaLibvirt()
        cpu_models = novalibvirt.cpu_models
        if cpu_models:
            _info["cpu-models"] = cpu_models

        vm_vcpu_info = novalibvirt.vcpu_info
        if vm_vcpu_info:
            _info["vcpu-info"] = vm_vcpu_info

        if _info:
            return _info

        return None


class SrcMigrationCallback(OpenstackEventCallbackBase):
    """
    Implements Openstack Nova live-migration source migration events
    callback.
    """
    event_group = 'nova.migrations'
    event_names = ['src-migration']

    @staticmethod
    def _migration_seq_info(event):
        """
        Process the results of an event that was defined as a sequence.

        Returns a dict keyed by a resource id inside which is a dict of
        information keyed by date. The date can optionally be extended to
        include the time to get greater granularity - requires the associated
        search patterns to include this information.

        @param event: EventCheckResult object
        """
        info = {}
        resource_idx = 3
        info_idxs = {'memory': 4, 'disk': 5}
        for section in event.results:
            for result in section:
                if not result.tag.endswith('-body'):
                    continue

                ts_date = result.get(1)
                ts_time = result.get(2)
                ts_date = f"{ts_date} {ts_time}"

                resource = result.get(resource_idx)
                if resource not in info:
                    info[resource] = {}

                if result.section_id not in info[resource]:
                    info[resource][result.section_id] = {}

                if ts_date not in info[resource]:
                    info[resource][result.section_id][ts_date] = {}

                resource_date = info[resource][result.section_id][ts_date]

                for name, idx in info_idxs.items():
                    value = result.get(idx)
                    if value is None:
                        continue

                    if name not in resource_date:
                        resource_date[name] = [value]
                    else:
                        resource_date[name].append(value)

        return info

    def _migration_vm_info(self, vm_uuid, start, end, duration):
        """
        Process the info from an instance

        @param vm_uuid: UUID of instance
        @param samples: Samples info of instance
        @param start: Start time of instance
        @param end: End time of instance
        @param duration: Duration time of instance
        """
        info = {'start': start, 'end': end, 'duration': duration}
        instance = self.project_helpers.nova.instances.get(vm_uuid)
        if instance and instance.memory_mbytes is not None:
            info['resources'] = {'memory_mbytes':
                                 instance.memory_mbytes}

        return info

    @staticmethod
    def _get_duration(start, end):
        _start = datetime.strptime(start, "%Y-%m-%d %H:%M:%S")
        _end = datetime.strptime(end, "%Y-%m-%d %H:%M:%S")
        return round(float((_end - _start).total_seconds()), 2)

    def __call__(self, event):
        """
        Source migration is defined as a sequence so that we can capture some
        of in the interim events such as memory and disk progress.
        """
        migration_info = {}

        results = self._migration_seq_info(event)
        for vm_uuid, sections in results.items():
            for section in sections.values():
                samples = {}
                start = None
                end = None
                for date, info in utils.sorted_dict(section).items():
                    if start is None:
                        start = date

                    end = date
                    for rtype, values in info.items():
                        if rtype not in samples:
                            samples[rtype] = []

                        samples[rtype] += [int(i) for i in values]

                duration = self._get_duration(start, end)
                info = self._migration_vm_info(vm_uuid, start, end, duration)
                if samples:
                    # regressions imply that the progress counter had one or
                    # more decreases before increasing again.
                    info['regressions'] = {}
                    for rtype, values in samples.items():
                        if 'iterations' not in info:
                            info['iterations'] = len(values)

                        info['regressions'][rtype] = (
                            utils.sample_set_regressions(values))

                if vm_uuid in migration_info:
                    migration_info[vm_uuid].append(info)
                else:
                    migration_info[vm_uuid] = [info]

        # section name expected to be live-migration
        return migration_info, event.section_name


class PrePostLiveMigrationCallback(OpenstackEventCallbackBase):
    """
    Implements Openstack Nova live-migration pre-post migration events
    callback.
    """
    event_group = 'nova.migrations'
    event_names = ['src-post-live-migration', 'dst-pre-live-migration']

    @staticmethod
    def _migration_stats_info(event):
        """
        Process events that have passthrough-results=True such that they can be
        passed directory to analytics.LogEventStats for parsing.
        """
        stats = LogEventStats(event.results, event.name)
        stats.run()
        top5 = stats.get_top_n_events_sorted(5)
        if not top5:
            return None

        results = {"top": top5}
        # There can be a very large number of incomplete migrations so need to
        # find a useful way to represent this
        # if stats.data.incomplete_events:
        #     results['incomplete-migrations'] = stats.data.incomplete_events

        return results

    def __call__(self, event):
        # section name expected to be live-migration
        return self._migration_stats_info(event), event.section_name


class NovaServerMigrationAnalysis(OpenstackEventHandlerBase):
    """ Implements Openstack Nova live-migration events handler. """
    event_group = 'nova.migrations'
    summary_part_index = 3

    @summary_entry('nova-migrations', get_min_available_entry_index() + 2)
    def summary_nova_migrations(self):
        return self.run()



from functools import cached_property

from hotsos.core.plugins.openstack.openstack import OSTServiceBase

OCTAVIA_HM_PORT_NAME = 'o-hm0'


class OctaviaBase(OSTServiceBase):
    """ Base class for Octavia checks. """
    PROJECT_NAME = 'octavia'

    @property
    def bind_interfaces(self):
        """
        Fetch interface o-hm0 used by Openstack Octavia. Returned dict is
        keyed by config key used to identify interface.
        """
        interfaces = {}
        port = self.nethelp.get_interface_with_name(OCTAVIA_HM_PORT_NAME)
        if port:
            interfaces.update({OCTAVIA_HM_PORT_NAME: port})

        return interfaces

    @property
    def hm_port_has_address(self):
        port = self.bind_interfaces.get(OCTAVIA_HM_PORT_NAME)
        if port is None or not port.addresses:
            return False

        return True

    @cached_property
    def hm_port_healthy(self):
        port = self.bind_interfaces.get(OCTAVIA_HM_PORT_NAME)
        if port is None:
            return True

        for counters in port.stats.values():
            total = sum(counters.values())
            if not total:
                continue

            pcent = int(100 / float(total) * float(counters.get('dropped', 0)))
            if pcent > 1:
                return False

            pcent = int(100 / float(total) * float(counters.get('errors', 0)))
            if pcent > 1:
                return False

        return True



from dataclasses import dataclass

# NOTE: this is date only no time
APACHE_DATE_EXPR = r'(\d{2}/\w{3,5}/\d{4})'
PYWSGI_DATE = r'([\d-]+) [\d:]+\.\d{3}'

HTTP_REQ_EXPR = r'"(GET|POST|PUT|DELETE)'
APACHE_REQ_COMMON: str = f'.+{APACHE_DATE_EXPR}.+ {HTTP_REQ_EXPR}'

WSGI_HTTP_STATUS_EXPR = r'HTTP/\d\.\d" status: (\d+)'
HTTP_STATUS_EXPR = r'HTTP/\d\.\d" (\d+)'
APACHE_STATUS_COMMON: str = f'.+{APACHE_DATE_EXPR}.+ {HTTP_STATUS_EXPR}'


@dataclass(frozen=True)
class HTTPRequestExprs:  # pylint: disable=too-many-instance-attributes
    """
    Search expressions for HTTP request type in OpenStack api logs.
    """
    # apache
    apache_common: str = APACHE_REQ_COMMON
    # python wsgi neutron
    wsgi_neutron: str = f'{PYWSGI_DATE} .+ neutron.wsgi .+{HTTP_REQ_EXPR}'
    # sunbeam
    sunbeam_cinder: str = r'.+ \[wsgi-cinder-api\]' + APACHE_REQ_COMMON
    sunbeam_glance: str = (r'.+ \[glance-api\] ' + PYWSGI_DATE +
                           f' .+ eventlet.wsgi.server .+ {HTTP_REQ_EXPR}')
    sunbeam_keystone: str = r'.+ \[wsgi-keystone\]' + APACHE_REQ_COMMON
    sunbeam_neutron: str = (f'.+ {PYWSGI_DATE} .+ neutron.wsgi .+'
                            f'{HTTP_REQ_EXPR}')
    sunbeam_nova: str = r'.+ \[wsgi-nova-api\]' + APACHE_REQ_COMMON
    sunbeam_octavia: str = r'.+ \[wsgi-octavia-api\]' + APACHE_REQ_COMMON
    sunbeam_placement: str = r'.+ \[wsgi-placement-api\]' + APACHE_REQ_COMMON


@dataclass(frozen=True)
class HTTPStatusExprs:  # pylint: disable=too-many-instance-attributes
    """
    Search expressions for HTTP return status in OpenStack api logs.
    """
    # apache
    apache_common: str = APACHE_STATUS_COMMON
    # python wsgi neutron
    wsgi_neutron: str = (r'([\d-]+) [\d:]+\.\d{3} .+ neutron.wsgi .+'
                         f'{WSGI_HTTP_STATUS_EXPR}')
    # sunbeam
    sunbeam_cinder: str = r'.+ \[wsgi-cinder-api\]' + APACHE_STATUS_COMMON
    sunbeam_glance: str = (r'.+ \[glance-api\] ' + PYWSGI_DATE +
                           f' .+ eventlet.wsgi.server .+ {HTTP_STATUS_EXPR}')
    sunbeam_keystone: str = r'.+ \[wsgi-keystone\]' + APACHE_STATUS_COMMON
    sunbeam_nova: str = r'.+ \[wsgi-nova-api\]' + APACHE_STATUS_COMMON
    sunbeam_neutron: str = (f'.+ {PYWSGI_DATE} .+ neutron.wsgi .+'
                            f'{WSGI_HTTP_STATUS_EXPR}')
    sunbeam_octavia: str = r'.+ \[wsgi-octavia-api\]' + APACHE_STATUS_COMMON
    sunbeam_placement: str = (r'.+ \[wsgi-placement-api\]' +
                              APACHE_STATUS_COMMON)



from hotsos.core.plugins.rabbitmq import RabbitMQChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class RabbitMQSummary(RabbitMQChecks):
    """ Implementation of RabbitMQ summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @property
    def queue_info(self):
        """Get distribution of queues across cluster."""
        vhosts = self.report.vhosts
        vhost_queue_dists = {}
        for vhost in vhosts:
            # include anyway to show empty vhosts
            vhost_queue_dists[vhost.name] = {}
            if not vhost.total_queues:
                continue

            for node, vhost_dist in vhost.node_queue_distributions.items():
                if vhost_dist['queues']:
                    vhost_queue_dists[vhost.name][node] = \
                        (f"{vhost_dist['queues']:d} "
                         f"({vhost_dist['pcent']:.2f}%)")
                else:
                    vhost_queue_dists[vhost.name][node] = \
                        f"{vhost_dist['queues']:d} (N/A)"

        _queue_info = {}
        if vhost_queue_dists:
            # list all vhosts but only show their queues if not []
            _queue_info["vhosts"] = sorted([vhost.name for vhost in vhosts])
            _queue_info["vhost-queue-distributions"] = \
                {k: v for k, v in vhost_queue_dists.items() if v}
            return _queue_info

        return None

    @summary_entry('config', get_min_available_entry_index())
    def summary_config(self):
        setting = self.report.partition_handling or 'unknown'
        return {'cluster-partition-handling': setting}

    @summary_entry('resources', get_min_available_entry_index() + 1)
    def summary_resources(self):
        resources = {}
        _queue_info = self.queue_info
        if _queue_info:
            resources.update(_queue_info)

        connections = self.report.connections
        if connections['host']:
            resources["connections-per-host"] = connections['host']

        if connections['client']:
            resources["client-connections"] = connections['client']

        if self.report.memory_used:
            resources["memory-used-mib"] = self.report.memory_used

        return resources or None



import re
from functools import cached_property

from hotsos.core.log import log
from hotsos.core.search import (
    create_constraint,
    SearchDef,
    SequenceSearchDef,
    ExtraSearchConstraints,
)
from hotsos.core.ycheck.engine.properties.common import (
    YPropertyOverrideBase,
    YPropertyMappedOverrideBase,
)
from hotsos.core.exceptions import (
    NotEnoughParametersError
)


class YPropertySearchConstraints(YPropertyOverrideBase):
    """
    Search constraints property member of search property.
    """
    override_keys = ['constraints']

    @property
    def valid_attributes(self):
        return ['search-period-hours', 'search-result-age-hours',
                'min-hours-since-last-boot', 'min-results']

    @cached_property
    def search_period_hours(self):
        """
        If min is provided this is used to determine the period within which
        min applies. If period is unset, the period is infinite i.e. across all
        available data.

        Supported values:
          <int> hours

        """
        return int(self.content.get('search-period-hours', 0))

    @cached_property
    def search_result_age_hours(self):
        """
        Result must have occurred within this number of hours from the current
        time (for a sosreport this would be when it was created).
        """
        return int(self.content.get('search-result-age-hours', 0))

    @cached_property
    def min_hours_since_last_boot(self):
        """
        Search result must be at least this number of hours after the last
        boot time.
        """
        return int(self.content.get('min-hours-since-last-boot', 0))

    @cached_property
    def min_results(self):
        """
        Minimum search matches required for result to be True (default is 1)
        """
        return int(self.content.get('min-results', 1))

    @cached_property
    def filesearch_constraints_obj(self):
        """
        Create a search constraints object representing the paramaters in
        this property.
        """
        invalid = []
        for attr in self.content:
            if attr not in self.valid_attributes:
                invalid.append(attr)

        if invalid:
            raise AttributeError(
                "Invalid search constraints attributes found: "
                f"{', '.join(invalid)}. "
                "Valid options are: "
                f"{', '.join(self.valid_attributes)}")
        return create_constraint(self.search_result_age_hours,
                                 self.min_hours_since_last_boot)


class YPropertySearchBase(YPropertyOverrideBase):
    """ Base class for search properties. """

    def apply_extra_constraints(self, results):
        """
        Apply further constraints filtering to search results.

        These are constraints supported by the constraints property that
        are not/cannot be applied by the search engine itself using
        SearchConstraintSearchSince.
        """
        if not self.constraints:
            log.debug("no extra search constraints to apply")
            return results

        log.debug("applying extra search constraints")
        count = len(results)
        results = ExtraSearchConstraints().apply(
                    results,
                    self.constraints.search_period_hours,
                    self.constraints.min_results)
        log.debug("applying extra search constraints reduced results from %s "
                  "to %s", count, len(results))
        return results

    @property
    def unique_search_tag(self):
        return self.override_path  # pylint: disable=E1101

    def _resolve_exprs(self, patterns):
        """
        Resolve any expressions provided as variable to their value. Non-vars
        are returned as-is.
        """
        _patterns = []
        if isinstance(patterns, list):
            for p in patterns:
                _patterns.append(self.resolve_var(p))  # pylint: disable=E1101
        else:
            _patterns.append(self.resolve_var(patterns))  # noqa, pylint: disable=E1101

        return _patterns

    @staticmethod
    def _is_import_path(value):
        return isinstance(value, str) and value.startswith('@')

    @property
    def search_pattern(self):
        try:
            if self.expr:
                if isinstance(self, YPropertySearch):
                    expr = self.expr.expr
                else:
                    expr = self.expr
            else:
                expr = self.content

            # can be supplied as a single string or list of strings
            patterns = self._resolve_exprs(expr)
            if len(patterns) == 0:
                raise NotEnoughParametersError(
                    "no search pattern (expr) defined")

            _patterns = []
            for pattern in patterns:
                if self._is_import_path(pattern):
                    try:
                        pattern = self.get_property(self.context, pattern[1:])
                    except ModuleNotFoundError:
                        log.warning("failed to import pattern '%s' - using "
                                    "raw string as expression", pattern)

                if not re.search(r"[^\\]?\(.+\)", pattern):
                    log.info("pattern '%s' does not contain a subgroup. this "
                             "is inefficient and can result in unnecessary "
                             "memory consumption", pattern)

                _patterns.append(pattern)

            if len(_patterns) == 1:
                return _patterns[0]

            return patterns
        except Exception:
            log.exception("")
            raise

    @property
    def is_sequence_search(self):
        seq_keys = YPropertySequencePart.get_override_keys_back_compat()
        return any(getattr(self, key) for key in seq_keys)

    @property
    def simple_search(self):
        if (self.is_sequence_search or not self.search_pattern or
                self.passthrough_results):
            return None

        sdef = self.cache.simple_search
        if sdef:
            return sdef

        pattern = self.search_pattern
        hint = None
        if self.hint:
            hint = str(self.hint)

        constraints = None
        if self.constraints and self.constraints.filesearch_constraints_obj:
            constraints = [self.constraints.filesearch_constraints_obj]

        sdef = SearchDef(pattern, tag=self.unique_search_tag, hint=hint,
                         constraints=constraints)

        self.cache.set('simple_search', sdef)
        return sdef

    @property
    def sequence_search(self):
        if not self.is_sequence_search or self.passthrough_results:
            return None

        sdef = self.cache.sequence_search
        if sdef:
            return sdef

        seq_start = self.start
        seq_body = self.body
        seq_end = self.end

        if (seq_body or (seq_end and not self.passthrough_results)):
            sd_start = SearchDef(seq_start.search_pattern)

            sd_end = None
            # explicit end is optional for sequence definition
            if seq_end:
                sd_end = SearchDef(seq_end.search_pattern)

            sd_body = None
            if seq_body:
                sd_body = SearchDef(seq_body.search_pattern)

            # NOTE: we don't use hints here
            tag = self.unique_search_tag
            sdef = SequenceSearchDef(start=sd_start, body=sd_body,
                                     end=sd_end, tag=tag)
            self.cache.set('sequence_search', sdef)
            return sdef

        log.warning("invalid sequence definition passthrough=%s "
                    "start=%s, body=%s, end=%s",
                    self.passthrough_results, seq_start, seq_body,
                    seq_end)
        return None

    @property
    def sequence_passthrough_search(self):
        if not self.is_sequence_search or not self.passthrough_results:
            return None

        sdef = self.cache.sequence_passthrough_search
        if sdef:
            return sdef

        seq_start = self.start
        seq_end = self.end

        if not (self.passthrough_results and all([seq_start, seq_end])):
            return None

        # start and end required for core.analytics.LogEventStats
        start_tag = f"{self.unique_search_tag}-start"
        end_tag = f"{self.unique_search_tag}-end"
        sdefs = [SearchDef(str(seq_start.search_pattern), tag=start_tag),
                 SearchDef(str(seq_end.search_pattern), tag=end_tag)]
        self.cache.set('sequence_passthrough_search', sdefs)
        return sdefs

    def load_searcher(self, searchobj, search_path, allow_constraints=True):
        """ Load search definitions into the given searcher object. """
        sdef = self.simple_search
        if sdef:
            log.debug("loading simple search")
            searchobj.add(sdef, search_path,
                          allow_global_constraints=allow_constraints)
        else:
            sdef = self.sequence_search
            if sdef:
                log.debug("loading sequence search")
                searchobj.add(sdef, search_path,
                              allow_global_constraints=allow_constraints)
            else:
                sdef = self.sequence_passthrough_search
                if sdef:
                    log.debug("loading sequence passthrough searches")
                    for _sdef in sdef:
                        searchobj.add(
                            _sdef,
                            search_path,
                            allow_global_constraints=allow_constraints)


class YPropertySearchOpt(YPropertyOverrideBase):
    """
    Provides search expression option properties that can be provided with each
    search definition.
    """
    override_keys = ['expr', 'hint', 'passthrough-results']

    def __bool__(self):
        return bool(self.content)

    @property
    def expr(self):
        """ Can be string or list. """
        return self.content

    def __str__(self):
        # should use bool() for passthrough-results
        invalid = ['passthrough-results', 'expr']
        valid = [k for k in
                 self.get_override_keys_back_compat() if k not in invalid]
        if self.override_name not in valid:
            raise AttributeError(
                f"__str__ only valid for {','.join(valid)} "
                f"(not {self.override_name})")

        return self.content


class SeqPartSearchOptsBase():
    """
    Provides implementation of extraction attributes to make it easy to get the
    value of properties supported by YPropertySearchOpt.
    """

    @property
    def expr(self):
        if not isinstance(self.content, dict):  # pylint: disable=E1101
            return self.content  # pylint: disable=E1101

        return self.content.get('expr', '')  # pylint: disable=E1101

    @property
    def hint(self):
        if not isinstance(self.content, dict):  # pylint: disable=E1101
            return None

        return self.content.get('hint', '')  # pylint: disable=E1101

    @property
    def passthrough_results(self):
        if not isinstance(self.content, dict):  # pylint: disable=E1101
            return False

        return self.content.get('passthrough-results', False)  # noqa, pylint: disable=E1101


class YPropertySequencePart(YPropertySearchBase, SeqPartSearchOptsBase,
                            YPropertyOverrideBase):
    """
    Provides the search fields required to perform a sequence search.
    """
    override_keys = ['start', 'body', 'end']


class YPropertySearch(YPropertySearchBase, YPropertyMappedOverrideBase):
    """ Search property. """
    override_keys = ['search']
    override_members = [YPropertySearchOpt, YPropertySearchConstraints,
                        YPropertySequencePart]

    @property
    def passthrough_results(self):
        """ Override the member to ensure we always return a bool. """
        if not isinstance(self.content, dict):
            return False

        return bool(self.content.get('passthrough-results', False))




# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' castellan/common/exception.py
CASTELLAN_EXCEPTIONS = [
    "RedirectException",
    "CastellanException",
    "Forbidden",
    "KeyManagerError",
    "ManagedObjectNotFoundError",
    "InvalidManagedObjectDictError",
    "UnknownManagedObjectTypeError",
    "AuthTypeInvalidError",
    "InsufficientCredentialDataError",
]




import errno
import glob
import os
import re
import subprocess
from functools import cached_property
from dataclasses import dataclass

import yaml
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core import utils


class JujuMachine():
    """ Juju machine interface. """
    def __init__(self, juju_lib_path):
        self.juju_lib_path = juju_lib_path

    @property
    def id(self):
        name = self.agent_service_name
        if name:
            return name.partition("jujud-machine-")[2]

        return None

    @cached_property
    def config(self):
        path = glob.glob(os.path.join(self.juju_lib_path,
                                      "agents/machine-*/agent.conf"))
        if not path:
            return {}

        # NOTE: we only expect one of these to exist
        path = path[0]
        if not os.path.exists(path):
            return None

        # filter out 'sanitised' lines since they will not be valid yaml
        ftmp = utils.mktemp_dump("")
        try:
            with open(ftmp, 'w', encoding='utf-8') as fdtmp:
                expr = re.compile(r"\*{9}")
                with open(path, encoding='utf-8') as fd:
                    for line in fd:
                        if not expr.search(line):
                            fdtmp.write(line)

            with open(ftmp, encoding='utf-8') as fd:
                cfg = yaml.safe_load(fd)
        finally:
            os.remove(ftmp)

        return cfg

    @cached_property
    def agent_service_name(self):
        return self.config.get("values", {}).get("AGENT_SERVICE_NAME")

    @property
    def agent_bin_path(self):
        """
        Return path to jujud machine binary.

        We could get this from systemd but for now this should work.
        """
        path = os.path.join(HotSOSConfig.data_root,
                            'var/lib/juju/tools/machine-*/jujud')
        for path in glob.glob(path):
            return path

        return None

    @cached_property
    def version(self):
        """
        Get the installed juju version.

        Juju may not have been installed by a package so we first try
        getting it from the binary then from config.
        """
        if self.agent_bin_path and os.path.exists(self.agent_bin_path):
            try:
                out = subprocess.check_output([self.agent_bin_path,
                                               '--version'])
                if not isinstance(out, str):
                    out = out.decode()

                return out.strip()
            except subprocess.CalledProcessError as exc:
                log.debug("failed to get juju version from %s (%s) - trying "
                          "config", self.agent_bin_path, exc)
            except OSError as exc:
                if exc.errno == errno.ENOEXEC:
                    # Probably a different arch than host where hotsos runs.
                    # So we get the version from the symlink.
                    _dir = os.readlink(os.path.dirname(self.agent_bin_path))
                    return os.path.basename(_dir)
                log.debug("failed to get juju version from %s (%s)",
                          self.agent_bin_path, exc.strerror)

        return self.config.get("upgradedToVersion", "unknown")

    @cached_property
    def deployed_units(self):
        units = []
        # requires >= 2.9.x
        _units = self.config.get("values", {}).get("deployed-units", "")
        if not _units:
            return units

        for unit in _units.split(','):
            app = unit.partition('/')[0]
            unit_id = unit.partition('/')[2]
            path = os.path.join(self.juju_lib_path,
                                f"agents/unit-{app}-{unit_id}")
            units.append(JujuUnit(unit_id, app, self.juju_lib_path, path=path))

        return units


class JujuUnit():
    """ Juju unit interface. """
    def __init__(self, unit_id, application, juju_lib_path, path=None):
        self.id = unit_id
        self.application = application
        self.name = f'{application}-{unit_id}'
        self.juju_lib_path = juju_lib_path
        self.path = path

    @cached_property
    def charm_name(self):
        """
        The deployer manifest file will give us the name of the charm used to
        deploy the unit whose name may not match the charm. It also tells us
        where the charm was deployed from i.e. cs:, ch: etc
        """
        manifest_path = f"agents/unit-{self.name}/state/deployer/manifests/*"
        for entry in glob.glob(os.path.join(self.juju_lib_path,
                                            manifest_path)):
            # we expect only one
            manifest_file = os.path.basename(entry)
            # e.g. ch_3a_amd64_2f_focal_2f_mysql-innodb-cluster-30
            return manifest_file.split('_')[-1].rpartition('-')[0]

        return None

    @cached_property
    def repo_info(self):
        """
        Some charms, e.g. the Openstack charms, provide a repo-info file that
        contains information from the charm's repository e.g. commit id.
        """
        info = {}
        path = os.path.join(self.path, 'charm/repo-info')
        if not os.path.exists(path):
            return info

        with open(path, encoding='utf-8') as fd:
            for line in fd:
                if line.startswith('commit-short:'):
                    sha1_short = line.partition(' ')[2]
                    if sha1_short:
                        info['commit'] = sha1_short.strip()

        return info


@dataclass(frozen=True)
class JujuCharm:
    """ Juju charm interface. """

    name: str
    version: int


class JujuBase():
    """ Juju checks base class. """
    CHARM_MANIFEST_GLOB = "agents/unit-*/state/deployer/manifests"

    @staticmethod
    def get_juju_lib_path():
        return os.path.join(HotSOSConfig.data_root, "var/lib/juju")

    @cached_property
    def machine(self):
        machine = JujuMachine(self.get_juju_lib_path())
        if not machine.config:
            log.debug("no juju machine identified")
            return None

        return machine

    @property
    def version(self):
        return self.machine.version

    @cached_property
    def units(self):
        """
        Returns units running on this host.

        @return: dict of JujuUnit objects keyed by unit name.
        """
        _units = {}
        if not os.path.exists(self.get_juju_lib_path()):
            return _units

        if self.machine and self.machine.version >= "2.9":
            _units = {u.name: u for u in self.machine.deployed_units}
        else:
            paths = glob.glob(os.path.join(self.get_juju_lib_path(),
                                           "agents/unit-*"))
            for unit in paths:
                base = os.path.basename(unit)
                ret = re.compile(r"unit-(\S+)-(\d+)").match(base)
                if ret:
                    app = ret.group(1)
                    unit_id = ret.group(2)
                    u = JujuUnit(unit_id, app, self.get_juju_lib_path(),
                                 path=unit)
                    _units[u.name] = u

        return _units

    @cached_property
    def charms(self):
        """
        Returns charms used by units on this host.

        @return: dict of JujuCharm objects keyed by charm name.
        """
        _charms = {}
        if not os.path.exists(self.get_juju_lib_path()):
            return _charms

        for entry in glob.glob(os.path.join(self.get_juju_lib_path(),
                                            self.CHARM_MANIFEST_GLOB)):
            name = None
            versions = []
            for manifest in os.listdir(entry):
                base = os.path.basename(manifest)
                ret = re.compile(r".+_(\S+)-(\d+)$").match(base)
                if ret:
                    name = ret.group(1)
                    versions.append(int(ret.group(2)))

            if name and versions:
                _charms[name] = JujuCharm(name, max(versions))

        return _charms

    @cached_property
    def charm_names(self):
        if not self.charms:
            return []

        return list(self.charms.keys())


class JujuBinaryInterface(JujuBase):
    """ Interface to juju binary. """
    @property
    def bin_names(self):
        return ['juju']

    def is_installed(self, name):
        return self.machine is not None and name in self.bin_names

    def get_version(self, _):
        return self.version



from hotsos.core.log import log
from hotsos.core.host_helpers import PebbleHelper
from hotsos.core.ycheck.engine.properties.requires import ServiceCheckItemsBase
from hotsos.core.ycheck.engine.properties.requires.types import (
    service_manager_common
)


class PebbleServiceCheckItems(ServiceCheckItemsBase):
    """
    Implementation for check items on pebble resource properties.
    """
    @property
    def _svcs_info(self):
        return PebbleHelper(self._svcs_all)


class YRequirementTypePebble(service_manager_common.ServiceManagerTypeBase):
    """
    Pebble requires type property. Provides support for defining checks on
    pebble service manager resources.
    """
    override_keys = ['pebble']
    override_autoregister = True
    default_op = 'eq'

    def _check_item_settings(self, svc_obj, settings, cache_info,
                             all_items):
        processes = None
        if isinstance(settings, str):
            state = settings
            ops = [[self.default_op, state]]
        else:
            processes = settings.get('processes')
            if processes:
                log.debug("checking systemd service processes: %s", processes)

            op = settings.get('op', self.default_op)
            if 'state' in settings:
                ops = [[op, settings.get('state')]]
            else:
                ops = []

        if processes and not all_items.processes_running(processes):
            log.debug("one or more pebble service processes not running "
                      "- %s", ', '.join(processes))
            return False

        cache_info[svc_obj.name]['ops'] = self.ops_to_str(ops)
        return self.apply_ops(ops, opinput=svc_obj.state)

    @property
    def items_to_check(self):
        return PebbleServiceCheckItems(self.content)



from hotsos.core.plugins.rabbitmq.common import (
    RabbitMQEventHandlerBase,
    RabbitMQEventCallbackBase,
)


class RabbitMQEventsCallback(RabbitMQEventCallbackBase):
    """ Callback for rabbitmq error events. """
    event_group = 'errors'
    event_names = ['connection-exception', 'delivery-ack-timeout',
                   'mnesia-error-event']

    def __call__(self, event):
        ret = self.categorise_events(event)
        return ret


class RabbitMQEventChecks(RabbitMQEventHandlerBase):
    """ Event handler for rabbitmq error events. """
    event_group = 'errors'
    summary_part_index = 1

    @property
    def summary_subkey(self):
        return 'rabbitmq-checks'



import glob
import os
from dataclasses import dataclass, field

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    PebbleHelper,
    SystemdHelper,
)
from hotsos.core import (
    host_helpers,
    plugintools,
)

SVC_VALID_SUFFIX = r'[0-9a-zA-Z-_]*'
MYSQL_SVC_EXPRS = [rf'mysql{SVC_VALID_SUFFIX}']
CORE_APT = ['mysql']


@dataclass
class MySQLInstallInfo(InstallInfoBase):
    """ MySQL installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=CORE_APT))
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(service_exprs=MYSQL_SVC_EXPRS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(
                                                service_exprs=MYSQL_SVC_EXPRS))


class MySQLChecks(plugintools.PluginPartBase):
    """ MySQL checks. """
    plugin_name = 'mysql'
    plugin_root_index = 3

    def __init__(self, *args, **kwargs):
        super().__init__()
        MySQLInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        return len(MySQLInstallInfo().apt.core) > 0


class MySQLConfig(host_helpers.IniConfigBase):
    """ MySQL config interface. """
    def __init__(self, *args, **kwargs):
        path = os.path.join(HotSOSConfig.data_root,
                            'etc/mysql/mysql.conf.d/mysqld.cnf')
        super().__init__(*args, path=path, **kwargs)


class MySQLRouterConfig(host_helpers.IniConfigBase):
    """ MySQL Router config interface. """
    def __init__(self, *args, **kwargs):
        path = os.path.join(HotSOSConfig.data_root,
                            'var/lib/mysql/*mysql-router/mysqlrouter.conf')
        # expect only one
        for f in glob.glob(path):
            path = f

        super().__init__(*args, path=path, **kwargs)



import abc
import datetime
import json
import os
import pathlib
import pickle
import tempfile
from dataclasses import dataclass, field
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.exceptions import (
    CLIExecError,
    SourceNotFound,
)
from hotsos.core.host_helpers.cli.common import (
    BinCmd,
    BinFileCmd,
    CmdOutput,
)
from hotsos.core.host_helpers.cli.catalog import CommandCatalog
from hotsos.core.host_helpers.common import HostHelpersBase
from hotsos.core.host_helpers.exceptions import CommandNotFound
from hotsos.core.log import log


class NullSource():
    """ Exception raised to indicate that a datasource is not available. """
    def __call__(self, *args, **kwargs):
        return CmdOutput([])


class JournalctlBase():
    """ Base class for journalctl command implementations. """
    @property
    def since_date(self):
        """
        Returns a string datetime to be used with journalctl --since. This time
        reflects the maximum depth of history we will search in the journal.

        The datetime value returned takes into account config from HotSOSConfig
        and has the format "YEAR-MONTH-DAY". It does not specify a time.
        """
        current = CLIHelper().date(format="--iso-8601")
        if not current:
            log.warning("could not determine since date for journalctl "
                        "command")
            return None

        ts = datetime.datetime.strptime(current, "%Y-%m-%d")
        if HotSOSConfig.use_all_logs:
            days = HotSOSConfig.max_logrotate_depth
        else:
            days = 1

        ts = ts - datetime.timedelta(days=days)
        return ts.strftime("%Y-%m-%d")


class JournalctlBinCmd(BinCmd, JournalctlBase):
    """ Implements binary journalctl command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.format_journalctl_cmd)

    def format_journalctl_cmd(self, **kwargs):
        """ Add optional extras to journalctl command. """
        if kwargs.get("opts"):
            self.cmd = f"{self.cmd} {kwargs.get('opts')}"

        if kwargs.get("unit"):
            self.cmd = f"{self.cmd} --unit {kwargs.get('unit')}"

        if kwargs.get("date"):
            self.cmd = f"{self.cmd} --since {kwargs.get('date')}"
        elif self.since_date:
            self.cmd = f"{self.cmd} --since {self.since_date}"


class JournalctlBinFileCmd(BinFileCmd, JournalctlBase):
    """ Implements file-based journalctl command.

    NOTE: this may suffer from incompatibility issues if the journal data was
          created with a different version of systemd-journald.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.preformat_sos_journalctl)

    def preformat_sos_journalctl(self, **kwargs):
        default_opts = '-oshort-iso'
        if kwargs.get("opts"):
            self.path = (f"journalctl {default_opts} {kwargs.get('opts')} "
                         f"-D {self.path}")
        else:
            self.path = f"journalctl {default_opts} -D {self.path}"

        if kwargs.get("unit"):
            self.path = f"{self.path} --unit {kwargs.get('unit')}"

        if kwargs.get("date"):
            self.path = f"{self.path} --since {kwargs.get('date')}"
        elif self.since_date:
            self.path = f"{self.path} --since {self.since_date}"


class CLICacheWrapper():
    """ Wrapper for cli cache. """
    def __init__(self, cache_load_f, cache_save_f):
        self.load_f = cache_load_f
        self.save_f = cache_save_f

    def load(self, key):
        return self.load_f(key)

    def save(self, key, value):
        return self.save_f(key, value)


@dataclass
class SourceRunner():
    """ Manager to control how we execute commands.

    Ensures that we try data sources in a consistent order.

    @param cmdkey: unique key identifying this command.
    @param sources: list of command source implementations.
    @param cache: CLICacheWrapper object.
    @param output_file: If a file path is provided the output of running a
                        command is saved to that file.
    @param catch_exceptions: By default we catch binary execution
                             exceptions and return an exit code rather than
                             allowing the exception to be raised. If not
                             required this can be set to False.
    """
    cmdkey: str
    sources: list
    cache: CLICacheWrapper
    output_file: str = field(default=None)
    catch_exceptions: bool = field(default=True)

    def __post_init__(self):
        # Command output can differ between CLIHelper and CLIHelperFile so we
        # need to cache them separately.
        if self.output_file:
            self.cache_cmdkey = f"{self.cmdkey}.file"
        else:
            self.cache_cmdkey = self.cmdkey

    def bsource(self, *args, **kwargs):
        # binary sources only apply if data_root is system root
        bin_out = None
        for bsource in [s for s in self.sources if s.TYPE == "BIN"]:
            cache = False
            # NOTE: we currently only support caching commands with no
            #       args.
            if not any([args, kwargs]):
                cache = True
                out = self.cache.load(self.cache_cmdkey)
                if out is not None:
                    return out

            try:
                if self.output_file:
                    # don't decode if we are going to be saving to a file
                    kwargs['skip_json_decode'] = True

                bin_out = bsource(*args, **kwargs)
                if cache and bin_out is not None:
                    try:
                        self.cache.save(self.cache_cmdkey, bin_out)
                    except pickle.PicklingError as exc:
                        log.info("unable to cache command '%s' output: %s",
                                 self.cmdkey, exc)

                # if command executed but returned nothing that still counts
                # as success.
                break
            except CLIExecError as exc:
                if not self.catch_exceptions:
                    raise

                bin_out = CmdOutput(exc.return_value)

        return bin_out

    def fsource(self, *args, **kwargs):
        for fsource in [s for s in self.sources if s.TYPE == "FILE"]:
            try:
                skip_load_contents = False
                if self.output_file:
                    skip_load_contents = True

                return fsource(*args, **kwargs,
                               skip_load_contents=skip_load_contents)
            except CLIExecError as exc:
                if not self.catch_exceptions:
                    raise

                return CmdOutput(exc.return_value)
            except SourceNotFound:
                pass

        return None

    def _execute(self, *args, **kwargs):
        # always try file sources first
        ret = self.fsource(*args, **kwargs)
        if ret is not None:
            return ret

        if HotSOSConfig.data_root != '/':
            return NullSource()()

        return self.bsource(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Execute the command using the appropriate source runner. These can be
        binary or file-based depending on whether data root points to / or a
        sosreport. File-based are attempted first.

        A command can have more than one source implementation so we must
        ensure they all have a chance to run.
        """
        out = self._execute(*args, **kwargs)
        if self.output_file:
            if out.source is not None:
                return out.source

            with open(self.output_file, 'w', encoding='utf-8') as fd:
                if isinstance(out.value, list):
                    fd.write(''.join(out.value))
                elif isinstance(out.value, dict):
                    fd.write(json.dumps(out.value))
                else:
                    fd.write(out.value)

                return self.output_file

        return out.value


class CLIHelperBase(HostHelpersBase):
    """ Base class for clihelper implementations. """
    def __init__(self):
        super().__init__()
        self.cli_cache = CLICacheWrapper(self.cache_load, self.cache_save)

    @property
    def cache_root(self):
        """ Cache at plugin level rather than globally. """
        return HotSOSConfig.plugin_tmp_dir

    @property
    def cache_type(self):
        return 'cli'

    @property
    def cache_name(self):
        return "commands"

    def cache_load(self, key):
        return self.cache.get(key)

    def cache_save(self, key, value):
        return self.cache.set(key, value)

    @cached_property
    def command_catalog(self):
        catalog = CommandCatalog()
        catalog.update({'journalctl':
                        [JournalctlBinCmd('journalctl -oshort-iso'),
                         JournalctlBinFileCmd('var/log/journal')]})
        return catalog

    @abc.abstractmethod
    def __getattr__(self, cmdname):
        """ This is how commands are run. The command is looked up in the
        catalog and it's runner object is returned. The caller is expetced to
        call() the returned object to execute the command.

        @param cmdname: name of command we want to execute. This must match a
        name used to register a handler in the catalog.
        @return: SourceRunner object.
        """


class CLIHelper(CLIHelperBase):
    """
    This is used when we want to have command output as the return value when
    a command is executed.
    """

    def __getattr__(self, cmdname):
        try:
            return SourceRunner(cmdname, self.command_catalog[cmdname],
                                self.cli_cache)
        except KeyError as exc:
            raise CommandNotFound(cmdname, exc) from exc


class CLIHelperFile(CLIHelperBase):
    """
    This is used when we want the return value of a command to be a path to a
    file containing the return value of executing that command.

    This will do one of two things; if the command output originates from a
    file e.g. a sosreport command output file, it will return the path to that
    file. If the command is executed as a binary, its output is written to a
    temporary file and the path to that file is returned.
    """

    def __init__(self, *args, catch_exceptions=True, delete_temp=True,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.delete_temp = delete_temp
        self._tmp_file_mtime = None
        self._catch_exceptions = catch_exceptions

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        do_delete = (self.delete_temp or
                     self._tmp_file_mtime ==
                     os.path.getmtime(self.output_file))
        if do_delete:
            os.remove(self.output_file)

        # We want exceptions to be raised
        return False

    @cached_property
    def output_file(self):
        path = tempfile.mktemp(dir=HotSOSConfig.plugin_tmp_dir)
        pathlib.Path(path).touch()
        self._tmp_file_mtime = os.path.getmtime(path)
        return path

    def __getattr__(self, cmdname):
        try:
            ret = SourceRunner(cmdname, self.command_catalog[cmdname],
                               self.cli_cache, output_file=self.output_file,
                               catch_exceptions=self._catch_exceptions)
            return ret
        except KeyError as exc:
            raise CommandNotFound(cmdname, exc) from exc

        return None



from collections import UserList

from hotsos.core.host_helpers.cli.common import BinCmd, FileCmd


OVNAliases = ['openstack-hypervisor.', 'microovn.']


class OVNDBCTLShowBinCmd(BinCmd):
    """ Implementation for binary ovndb ctl show commands providing support
        for different variants.
    """

    def __init__(self, ctl_command, *args, prefix=None, **kwargs):
        prefix = prefix or ''
        cmd = f'{prefix}{ctl_command} --no-leader-only show'
        super().__init__(cmd, *args, **kwargs)


class OVNDBCTLShowFileCmd(FileCmd):
    """ Implementation for file-based ovndb ctl show commands providing support
        for different variants.
    """

    def __init__(self, ctl_command, *args, prefix=None, **kwargs):
        prefix = prefix or ''
        path = (f'sos_commands/ovn_central/{prefix}{ctl_command}_'
                '--no-leader-only_show')
        super().__init__(path, *args, **kwargs)


class OVNDBCTLShowCmds(UserList):
    """ Generate ovn-nbctl show command variants. """

    def __init__(self, ctl_command):
        cmds = [OVNDBCTLShowBinCmd(ctl_command=ctl_command)]
        cmds.extend([OVNDBCTLShowBinCmd(ctl_command=ctl_command,
                     prefix=prefix) for prefix in OVNAliases])
        cmds.append(OVNDBCTLShowFileCmd(ctl_command=ctl_command))
        # sosreport < 4.5
        cmds.append(FileCmd(f'sos_commands/ovn_central/{ctl_command}_show'))
        cmds.extend([OVNDBCTLShowFileCmd(ctl_command=ctl_command,
                     prefix=prefix) for prefix in OVNAliases])
        super().__init__(cmds)


class OVNDBCTLListBinCmd(BinCmd):
    """ Implementation for binary ovndb ctl list commands providing support
        for different variants.
    """

    def __init__(self, ctl_command, *args, prefix=None, **kwargs):
        prefix = prefix or ''
        cmd = f'{prefix}{ctl_command} --no-leader-only list' + ' {table}'
        super().__init__(cmd, *args, **kwargs)


class OVNDBCTLListFileCmd(FileCmd):
    """ Implementation for file-based ovndb ctl list commands providing support
        for different variants.
    """

    def __init__(self, ctl_command, *args, prefix=None, **kwargs):
        prefix = prefix or ''
        path = (f'sos_commands/ovn_central/{prefix}{ctl_command}_'
                '--no-leader-only_list_{table}')
        super().__init__(path, *args, **kwargs)


class OVNDBCTLListCmds(UserList):
    """ Generate ovn-nbctl list command variants. """

    def __init__(self, ctl_command):
        cmds = [OVNDBCTLListBinCmd(ctl_command=ctl_command)]
        cmds.extend([OVNDBCTLListBinCmd(ctl_command=ctl_command,
                     prefix=prefix) for prefix in OVNAliases])
        cmds.append(OVNDBCTLListFileCmd(ctl_command=ctl_command))
        # sosreport < 4.5
        cmds.append(FileCmd(f'sos_commands/ovn_central/{ctl_command}_list'))
        cmds.extend([OVNDBCTLListFileCmd(ctl_command=ctl_command,
                     prefix=prefix) for prefix in OVNAliases])
        super().__init__(cmds)



import re
from functools import cached_property

from hotsos.core.host_helpers import CLIHelper
from hotsos.core.log import log


class UptimeHelper():
    """ Helper to enable querying system uptime. """
    def __init__(self):
        """
        uptime is either expressed in different combinations of days, hours
        and minutes so we determine which format is used to then extract
        the time to present it in a common way.
        """
        # unfortunately sosreports don't have proc/uptime otherwise we would
        # use that.
        self.uptime = CLIHelper().uptime() or ""
        # this needs to take into account the different formats supported by
        # https://gitlab.com/procps-ng/procps/-/blob/newlib/library/uptime.c
        etime_expr = r"(?:([\d:]+)|(\d+\s+\S+,\s+[\d:]+)|(\d+\s+\S+)),"
        expr = rf"\s*[\d:]+\s+up\s+{etime_expr}.+\s+load average:\s+(.+)"
        ret = re.compile(expr).match(self.uptime)
        self.subgroups = {}
        if ret:
            self.subgroups['hour'] = {'value': ret.group(1),
                                      'expr': r'(\d+):(\d+)'}
            self.subgroups['day'] = {'value': ret.group(2),
                                     'expr': r"(\d+)\s+\S+,\s+(\d+):(\d+)"}
            self.subgroups['min'] = {'value': ret.group(3),
                                     'expr': r"(\d+)\s+(\S+)"}
            self.subgroups['loadavg'] = {'value': ret.group(4)}

    @cached_property
    def in_minutes(self):
        """ Total uptime in minutes. """
        if not self.subgroups:
            log.warning("uptime not available")
            return 0

        if self.subgroups['hour']['value']:
            expr = self.subgroups['hour']['expr']
            ret = re.match(expr, self.subgroups['hour']['value'])
            if ret:
                return (int(ret.group(1)) * 60) + int(ret.group(2))
        elif self.subgroups['day']['value']:
            expr = self.subgroups['day']['expr']
            ret = re.match(expr, self.subgroups['day']['value'])
            if ret:
                count = int(ret.group(1))
                hours = int(ret.group(2))
                mins = int(ret.group(3))
                day_mins = 24 * 60
                total = count * day_mins
                total += hours * 60
                total += mins
                return total
        elif self.subgroups['min']['value']:
            expr = self.subgroups['min']['expr']
            ret = re.match(expr, self.subgroups['min']['value'])
            if ret:
                return int(ret.group(1))

        log.warning("unknown uptime format in %s", self.uptime)
        return 0

    @property
    def in_seconds(self):
        """ Total uptime in seconds. """
        if self.in_minutes:
            return self.in_minutes * 60

        return 0

    @property
    def in_hours(self):
        """ Total uptime in hours. """
        if self.in_minutes:
            return int(self.in_minutes / 60)

        return 0

    def __repr__(self):
        days = int(self.in_hours / 24)
        hours = self.in_hours - days * 24
        minutes = self.in_minutes - (self.in_hours * 60)
        return f"{days}d:{hours}h:{minutes}m"

    @property
    def loadavg(self):
        if self.subgroups:
            return self.subgroups['loadavg']['value']

        return None



from dataclasses import dataclass, field

from hotsos.core.plugintools import PluginPartBase
from hotsos.core.host_helpers import (
    InstallInfoBase,
    PebbleHelper,
    SnapPackageHelper,
    SystemdHelper,
)

CORE_SNAPS = ['vault']
SERVICE_EXPRS = [s + '[A-Za-z0-9-]*' for s in CORE_SNAPS]


@dataclass
class VaultInstallInfo(InstallInfoBase):
    """ Vault installation information. """
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(service_exprs=SERVICE_EXPRS))
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(core_snaps=CORE_SNAPS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=SERVICE_EXPRS))


class VaultChecks(PluginPartBase):
    """ Base class for all vault checks. """
    plugin_name = 'vault'
    plugin_root_index = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        VaultInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        if VaultInstallInfo().snaps.core:
            return True

        return False



import re
from functools import cached_property

from hotsos.core.host_helpers import (
    CLIHelper,
    CLIHelperFile,
)
from hotsos.core.search import (
    FileSearcher,
    SequenceSearchDef,
    SearchDef
)
from hotsos.core.plugins.storage.ceph.daemon import (
    CephMon,
    CephOSD,
)
from hotsos.core.utils import sorted_dict

CEPH_POOL_TYPE = {1: 'replicated', 3: 'erasure-coded'}


class CephCrushMap():
    """
    Representation of a Ceph cluster CRUSH map.
    """
    @staticmethod
    def _filter_pools_by_rule(pools, crush_rule):
        res_pool = []
        for pool in pools:
            if pool['crush_rule'] == crush_rule:
                pool_str = pool['pool_name'] + ' (' + str(pool['pool']) + ')'
                res_pool.append(pool_str)

        return res_pool

    @cached_property
    def osd_crush_dump(self):
        return CLIHelper().ceph_osd_crush_dump_json_decoded() or {}

    @cached_property
    def ceph_report(self):
        return CLIHelper().ceph_report_json_decoded() or {}

    @cached_property
    def rules(self):
        """
        Returns a list of crush rules, mapped to the respective pools.
        """
        if not self.ceph_report:
            return {}

        rule_to_pool = {}
        for rule in self.ceph_report['crushmap']['rules']:
            rule_id = rule['rule_id']
            rtype = rule['type']
            pools = self.ceph_report['osdmap']['pools']
            pools = self._filter_pools_by_rule(pools, rule_id)
            rule_to_pool[rule['rule_name']] = {'id': rule_id,
                                               'type': CEPH_POOL_TYPE[rtype],
                                               'pools': pools}

        return rule_to_pool

    @staticmethod
    def _build_buckets_from_crushdump(crushdump):
        buckets = {}
        # iterate jp for each bucket
        for bucket in crushdump["buckets"]:
            bid = bucket["id"]
            items = []
            for item in bucket["items"]:
                items.append(item["id"])

            buckets[bid] = {"name": bucket["name"],
                            "type_id": bucket["type_id"],
                            "type_name": bucket["type_name"],
                            "items": items}

        return buckets

    def _rule_used_by_any_pool(self, rule_id):
        for pool_dict in self.rules.values():
            if (pool_dict['id'] == rule_id) and pool_dict['pools']:
                return True
        return False

    @cached_property
    def crushmap_mixed_buckets(self):
        """
        Report buckets that have mixed type of items,
        as they will cause crush map unable to compute
        the expected up set
        """
        if not self.osd_crush_dump:
            return []

        bad_buckets = []
        buckets = self._build_buckets_from_crushdump(self.osd_crush_dump)
        # check all buckets
        for bdict in buckets.values():
            items = bdict["items"]
            type_ids = []
            for item in items:
                if item >= 0:
                    type_ids.append(0)
                else:
                    type_ids.append(buckets[item]["type_id"])

            if not type_ids:
                continue

            # verify if the type_id list contain mixed type id
            if type_ids.count(type_ids[0]) != len(type_ids):
                bad_buckets.append(bdict["name"])

        return bad_buckets

    @cached_property
    def crushmap_mixed_buckets_str(self):
        return ','.join(self.crushmap_mixed_buckets)

    def _is_bucket_imbalanced(self, buckets, start_bucket_id, failure_domain,
                              weight=-1):
        """Return whether a tree is unbalanced

        Recursively determine if a given tree (start_bucket_id) is
        balanced at the given failure domain (failure_domain) in the
        CRUSH tree(s) provided by the buckets parameter.
        """

        for item in buckets[start_bucket_id]["items"]:
            if buckets[item["id"]]["type_name"] != failure_domain:
                if self._is_bucket_imbalanced(buckets, item["id"],
                                              failure_domain, weight):
                    return True
            # Handle items/buckets with 0 weight correctly, by
            # ignoring them.
            # These are excluded from placement consideration,
            # and therefore do not unbalance a tree.
            elif item["weight"] > 0:
                if weight == -1:
                    weight = item["weight"]
                else:
                    if weight != item["weight"]:
                        return True

        return False

    @cached_property
    def crushmap_equal_buckets(self):
        """
        Report when in-use failure domain buckets are unbalanced.

        Uses the trees and failure domains referenced in the
        CRUSH rules, and checks that all buckets of the failure
        domain type in the referenced tree are equal or of zero size.
        """
        if not self.osd_crush_dump:
            return []

        buckets = {b['id']: b for b in self.osd_crush_dump["buckets"]}

        to_check = []
        for rule in self.osd_crush_dump.get('rules', []):
            taken = 0
            fdomain = 0
            rid = rule["rule_id"]
            for i in rule['steps']:
                if i["op"] == "take":
                    taken = i["item"]
                if "type" in i and taken != 0:
                    fdomain = i["type"]
                if taken != 0 and fdomain != 0 and \
                        self._rule_used_by_any_pool(rid):
                    to_check.append((rid, taken, fdomain))
                    taken = fdomain = 0

        unequal_buckets = []
        for _, tree, failure_domain in to_check:
            if self._is_bucket_imbalanced(buckets, tree, failure_domain):
                unequal_buckets.append(f"tree '{buckets[tree]['name']}' at "
                                       f"the '{failure_domain}' level")

        return unequal_buckets

    @cached_property
    def crushmap_equal_buckets_pretty(self):
        unequal = self.crushmap_equal_buckets
        if unequal:
            return ", ".join(unequal)

        return None

    @cached_property
    def autoscaler_enabled_pools(self):
        if not self.ceph_report:
            return []

        pools = self.ceph_report['osdmap']['pools']
        return [p for p in pools if p.get('pg_autoscale_mode') == 'on']

    @cached_property
    def autoscaler_disabled_pools(self):
        if not self.ceph_report:
            return []

        pools = self.ceph_report['osdmap']['pools']
        return [p for p in pools if p.get('pg_autoscale_mode') != 'on']

    @cached_property
    def is_rgw_using_civetweb(self):
        if not self.ceph_report:
            return []

        try:
            rgws = self.ceph_report['servicemap']['services']['rgw']['daemons']
            for _, outer_d in rgws.items():
                if isinstance(outer_d, dict):
                    if outer_d['metadata']['frontend_type#0'] == 'civetweb':
                        return True
        except (ValueError, KeyError):
            pass

        return False


class CephCluster():  # pylint: disable=too-many-public-methods
    """
    Provides an interface to a Ceph cluster.

    NOTE: we disable the too-many-public-methods pylint warning because by
          nature this is going to have a large number of public methods.
    """
    OSD_META_LIMIT_PERCENT = 5
    OSD_PG_MAX_LIMIT = 500
    # min cluster utilisation required to trigger this warning
    OSD_PG_OPTIMAL_MIN_UTIL = 10
    OSD_PG_OPTIMAL_NUM_MAX = 200
    OSD_PG_OPTIMAL_NUM_MIN = 50
    OSD_DISCREPANCY_ALLOWED = 5
    # If a pool's utilisation is below this value, we consider it "empty"
    POOL_EMPTY_THRESHOLD = 2

    def __init__(self):
        self.crush_map = CephCrushMap()

    @cached_property
    def health_status(self):
        status = CLIHelper().ceph_status_json_decoded()
        if status:
            return status['health']['status']

        return None

    @cached_property
    def _mon_dump(self):
        return CLIHelper().ceph_mon_dump_json_decoded() or {}

    @cached_property
    def _osd_dump(self):
        return CLIHelper().ceph_osd_dump_json_decoded() or {}

    @cached_property
    def pg_dump(self):
        return CLIHelper().ceph_pg_dump_json_decoded() or {}

    @cached_property
    def _ceph_mgr_module_ls(self):
        return CLIHelper().ceph_mgr_module_ls() or {}

    @cached_property
    def mons(self):
        _mons = []
        for mon in self._mon_dump.get('mons', {}):
            _mons.append(CephMon(mon['name']))

        return _mons

    @cached_property
    def mgr_modules(self):
        """
        Returns a list of modules that are enabled. This includes both
        the 'always on' as well as modules enabled explicitly.
        """
        if not self._ceph_mgr_module_ls:
            return []

        _modules = []
        for category in ['always_on_modules', 'enabled_modules']:
            if self._ceph_mgr_module_ls[category]:
                _modules += self._ceph_mgr_module_ls[category]

        return _modules

    @cached_property
    def osd_df_tree(self):
        return CLIHelper().ceph_osd_df_tree_json_decoded() or {}

    @cached_property
    def ceph_df(self):
        return CLIHelper().ceph_df_json_decoded() or {}

    @cached_property
    def osds(self):
        """ Returns a list of CephOSD objects for all osds in the cluster. """
        _osds = []
        for osd in self._osd_dump.get('osds', {}):
            _osds.append(CephOSD(osd['osd'], osd['uuid'], dump=osd))

        return _osds

    @cached_property
    def nearfull_ratio(self):
        """Return the cluster nearfull ratio from ``ceph report``.

        Returns the value if available, otherwise ``None``.
        """
        report = self.crush_map.ceph_report
        if not report:
            return None

        val = None
        if 'osdmap' in report:
            val = report['osdmap'].get('nearfull_ratio')
        if val is None:
            val = report.get('nearfull_ratio')

        try:
            return round(val, 2) if val is not None else None
        except (TypeError, ValueError):
            return None

    @cached_property
    def backfillfull_ratio(self):
        """Return the cluster backfillfull ratio from ``ceph report``.

        Returns the value if available, otherwise ``None``.
        """
        report = self.crush_map.ceph_report
        if not report:
            return None

        val = None
        if 'osdmap' in report:
            val = report['osdmap'].get('backfillfull_ratio')
        if val is None:
            val = report.get('backfillfull_ratio')

        try:
            return round(val, 2) if val is not None else None
        except (TypeError, ValueError):
            return None

    @cached_property
    def full_ratio(self):
        """Return the cluster full ratio from ``ceph report``.

        Returns the value if available, otherwise ``None``.
        """
        report = self.crush_map.ceph_report
        if not report:
            return None

        val = None
        if 'osdmap' in report:
            val = report['osdmap'].get('full_ratio')
        if val is None:
            val = report.get('full_ratio')

        try:
            return round(val, 2) if val is not None else None
        except (TypeError, ValueError):
            return None

    @cached_property
    def cluster_osds_without_v2_messenger_protocol(self):
        v1_osds = []
        for osd in self.osds:
            for key, val in osd.dump.items():
                if key.endswith('_addrs'):
                    vers = [info['type'] for info in val.get('addrvec', [])]
                    if 'v2' not in vers:
                        v1_osds.append(osd.id)

        return v1_osds

    @staticmethod
    def _get_version_info(daemon_type=None):
        """
        Returns a dict of ceph versions info for the provided daemon type. If
        no daemon type provided, version info is collected for all types and
        the resulting dict is keyed by daemon type otherwise it is keyed by
        version (and only versions for that daemon type.)
        """
        versions = {}
        s = FileSearcher()
        body = SearchDef(r"\s+\"ceph version (\S+) .+ (\S+) "
                         r"\(\S+\)\":\s+(\d)+,?$")
        if daemon_type is None:
            # all/any - start matches any so no seq ending needed
            sd = SequenceSearchDef(start=SearchDef(r"^\s+\"(\S+)\":\s+{"),
                                   body=body, tag='versions')
        else:
            start = SearchDef(rf"^\s+\"({daemon_type})\":\s+{{")
            sd = SequenceSearchDef(start=start, body=body,
                                   end=SearchDef(r"^\s+\"\S+\":\s+{"),
                                   tag='versions')

        with CLIHelperFile() as cli:
            s.add(sd, path=cli.ceph_versions())
            for section in s.run().find_sequence_sections(sd).values():
                _versions = {}
                for result in section:
                    if result.tag == sd.start_tag:
                        _daemon_type = result.get(1)
                        versions[_daemon_type] = _versions
                    elif result.tag == sd.body_tag:
                        version = result.get(1)
                        rname = result.get(2)
                        amount = result.get(3)
                        _versions[version] = {'release_name': rname,
                                              'count': int(amount)}

        # If specific daemon_type provided only return version for that type
        # otherwise all.
        if daemon_type is not None:
            versions = versions.get(daemon_type)

        return versions

    def daemon_versions(self, daemon_type=None):
        """
        Returns a dict of versions of daemon type associated with the
        number of each that is running. Ideally this would only return a single
        version showing that all daemons are in sync but sometimes e.g.
        during an upgrade this may not be the case.
        """
        _versions = {}
        version_info = self._get_version_info(daemon_type)
        if version_info:
            if daemon_type:
                for ver, info in version_info.items():
                    _versions[ver] = info['count']
            else:
                for daemon, _version_info in version_info.items():
                    for ver, info in _version_info.items():
                        if daemon not in _versions:
                            _versions[daemon] = {}

                        _versions[daemon][ver] = info['count']

        return _versions

    def daemon_release_names(self, daemon_type=None):
        """
        Same as versions property but with release names instead of versions.
        """
        _releases = {}
        version_info = self._get_version_info(daemon_type)
        if version_info:
            if daemon_type:
                for info in version_info.values():
                    rname = info['release_name']
                    if rname in _releases:
                        _releases[rname] += info['count']
                    else:
                        _releases[rname] = info['count']
            else:
                for daemon, _version_info in version_info.items():
                    for info in _version_info.values():
                        if daemon not in _releases:
                            _releases[daemon] = {}

                        rname = info['release_name']
                        if rname in _releases:
                            _releases[daemon][rname] += info['count']
                        else:
                            _releases[daemon][rname] = info['count']

        return _releases

    @cached_property
    def osd_release_names(self):
        """
        Same as versions property but with release names instead of versions.
        """
        return self.daemon_release_names('osd')

    @cached_property
    def require_osd_release(self):
        return self._osd_dump.get('require_osd_release')

    @cached_property
    def osd_daemon_release_names_match_required(self):
        """
        Does the cluster have require_osd_release set to a specific release
        name and if so, do all osds match that release name.
        """
        required_rname = self.require_osd_release
        if not required_rname:
            return True

        rnames = set(self.daemon_release_names('osd').keys())
        return len(rnames) == 1 and required_rname in rnames

    @cached_property
    def laggy_pgs(self):
        if not self.pg_dump:
            return []

        laggy_pgs = []
        # The states we consider problematic for network related issues
        laggy_like_states = ['laggy', 'wait']
        for pg in self.pg_dump['pg_map']['pg_stats']:
            if any(x in pg['state'].split('+') for x in laggy_like_states):
                laggy_pgs.append(pg)

        return laggy_pgs

    def pool_id_to_name(self, pool_id):
        if not self._osd_dump:
            return None

        pools = self._osd_dump.get('pools', [])
        for pool in pools:
            if pool['pool'] == int(pool_id):
                return pool['pool_name']

        return None

    @cached_property
    def large_omap_pgs(self):
        _large_omap_pgs = {}
        if not self.pg_dump:
            return _large_omap_pgs

        for pg in self.pg_dump['pg_map']['pg_stats']:
            if pg['stat_sum']['num_large_omap_objects'] > 0:
                pg_id = pg['pgid']
                pg_pool_id = pg_id.partition('.')[0]
                _large_omap_pgs[pg['pgid']] = {
                    'pool': self.pool_id_to_name(pg_pool_id),
                    'last_scrub_stamp': pg['last_scrub_stamp'],
                    'last_deep_scrub_stamp': pg['last_deep_scrub_stamp']
                }

        return _large_omap_pgs

    @cached_property
    def large_omap_pgs_str(self):
        if not self.large_omap_pgs:
            return None

        return ', '.join(self.large_omap_pgs.keys())

    @cached_property
    def bluefs_oversized_metadata_osds(self):
        _bad_meta_osds = []
        if not self.osd_df_tree:
            return _bad_meta_osds

        for device in self.osd_df_tree['nodes']:
            if device['id'] >= 0:
                meta_kb = device['kb_used_meta']
                # Util under under 2G are ignored (treated as non-issue).
                # See #434 for relevant info.
                if meta_kb < (2 * 1024 * 1024):
                    continue
                total_kb = device['kb_used']
                if meta_kb > (self.OSD_META_LIMIT_PERCENT / 100.0 * total_kb):
                    _bad_meta_osds.append(device['name'])

        return sorted(_bad_meta_osds)

    @cached_property
    def cluster_has_non_empty_pools(self):
        """
        Although we say "non-empty", we don't actually expect pools to have
        0% utilisation in practice because:
          - there may test pools after a fresh deployment done by FE/users
          - device_health_metrics pool created automatically
        So we consider that pools empty if the usage is below a threshold.
        """
        if not self.ceph_df:
            return False

        for pool in self.ceph_df['pools']:
            if pool['stats']['percent_used'] > \
                    self.POOL_EMPTY_THRESHOLD / 100.0:
                return True
        return False

    @staticmethod
    def version_as_a_tuple(ver):
        """
        Converts dotted version strings to a tuple so that they can be
        compared with standard operators.

        For example 1.2.3 becomes (1, 2, 3).
        """
        return tuple(map(int, (ver.split("."))))

    def ceph_daemon_versions_unique(self, exclude_daemons=None):
        """
        Returns dict of unique daemon versions from the Cluster. In an ideal
        world all versions would be the same but it is common for them to get
        out of sync as components are upgraded asynchronously.
        """
        versions = self.daemon_versions()
        if not versions:
            return {}

        unique_versions = {}
        for daemon_type, count in versions.items():
            # skip the catchall
            if daemon_type == 'overall':
                continue

            if exclude_daemons:
                if daemon_type in exclude_daemons:
                    continue

            unique_versions[daemon_type] = sorted(
                list(set(count)))

        return unique_versions

    @cached_property
    def ceph_versions_aligned(self):
        versions = self.ceph_daemon_versions_unique()
        if not versions:
            return True

        gloval_vers = set()
        for vers in versions.values():
            if len(vers) > 1:
                return False

            gloval_vers.add(vers[0])

        return len(gloval_vers) == 1

    @cached_property
    def mon_versions_aligned_with_cluster(self):
        """
        While it is not critical to have minor version mismatches across
        daemons in the cluster, it is required for the Ceph mons to always
        be >= to the rest of the cluster. This returns True if the requirement
        is met otherwise False.
        """
        non_mon = self.ceph_daemon_versions_unique(exclude_daemons='mon')
        mon_versions = self.daemon_versions('mon')
        if not non_mon or not mon_versions:
            return True

        mon_top = sorted(mon_versions.keys(),
                         key=self.version_as_a_tuple)[0]
        for vers in non_mon.values():
            cl_top = sorted(vers, key=self.version_as_a_tuple)[0]
            if cl_top > mon_top:
                return False

        return True

    @cached_property
    def osdmaps_count(self):
        report = CLIHelper().ceph_report_json_decoded()
        if not report:
            return 0

        try:
            return len(report['osdmap_manifest']['pinned_maps'])
        except (ValueError, KeyError):
            return 0

    @cached_property
    def osds_pgs(self):
        _osds_pgs = {}
        if not self.osd_df_tree:
            return _osds_pgs

        for device in self.osd_df_tree['nodes']:
            if device['id'] >= 0:
                _osds_pgs[device['name']] = device['pgs']

        return _osds_pgs

    @cached_property
    def osds_pgs_above_max(self):
        _osds_pgs = {}
        for osd, num_pgs in self.osds_pgs.items():
            if num_pgs > self.OSD_PG_MAX_LIMIT:
                _osds_pgs[osd] = num_pgs

        return sorted_dict(_osds_pgs, key=lambda e: e[1], reverse=True)

    @cached_property
    def osds_pgs_suboptimal(self):
        _osds_pgs = {}
        if not self.osd_df_tree:
            return _osds_pgs

        total_util = self.osd_df_tree['summary']['average_utilization']
        # If the overall cluster utilisation is lower than 10%, we
        # skip this warning as this is likely to be non-issue due to
        # the cluster being near-empty where the PG count can't be
        # increased (unless done manually which isn't recommended).
        # The PG autoscaler will adjust the PGs (split or merge) as
        # the cluster grows in size.
        if total_util < self.OSD_PG_OPTIMAL_MIN_UTIL:
            return _osds_pgs

        for osd, num_pgs in self.osds_pgs.items():
            # allow 30% margin from optimal OSD_PG_OPTIMAL_NUM_* values
            margin_high = self.OSD_PG_OPTIMAL_NUM_MAX * 1.3
            margin_low = self.OSD_PG_OPTIMAL_NUM_MIN * .7
            if margin_high < num_pgs or margin_low > num_pgs:
                _osds_pgs[osd] = num_pgs

        return sorted_dict(_osds_pgs, key=lambda e: e[1], reverse=True)

    @cached_property
    def ssds_using_bcache(self):
        report = CLIHelper().ceph_report_json_decoded()
        if not report:
            return []

        ssd_osds_using_bcache = []
        for osd in report['osd_metadata']:
            if osd['bluestore_bdev_type'] == 'ssd' and \
                    osd['bluestore_bdev_rotational'] == '0' and \
                    re.search("bcache", osd['bluestore_bdev_devices']):
                ssd_osds_using_bcache.append(osd['id'])

        return sorted(ssd_osds_using_bcache)

    @staticmethod
    def _get_db_size_of_osd(osd_id):
        """
        Returns the bluefs DB size of the given OSD.
        Returned size is in bytes.
        """

        report = CLIHelper().ceph_report_json_decoded()
        if report:
            for osd in report['osd_metadata']:
                if osd['id'] == osd_id:
                    try:
                        return int(osd['bluefs_db_size'])
                    except KeyError:
                        # older versions do not output bluefs_db_size
                        return 0

        return 0

    @cached_property
    def osd_raw_usage_higher_than_data(self):
        _bad_osds = []

        if not self.osd_df_tree:
            return _bad_osds

        for osd in self.osd_df_tree['nodes']:
            osd_id = osd['id']
            if osd_id >= 0:
                raw_usage = osd['kb_used']
                db_size_kb = self._get_db_size_of_osd(osd_id) / 1024.0
                total_usage = osd['kb_used_data'] + osd['kb_used_omap'] + \
                    osd['kb_used_meta'] + db_size_kb
                # There's always some additional space used by OSDs that's not
                # by data/omap/meta for journaling, internal structures, etc.
                # Thus we allow 5% discrepancy.
                allowance = total_usage * self.OSD_DISCREPANCY_ALLOWED / 100.0
                if raw_usage > (total_usage + allowance):
                    _bad_osds.append(osd['name'])

        return sorted(_bad_osds)



from hotsos.core.host_helpers import NetworkPort
from hotsos.core.plugins.openvswitch import OpenvSwitchChecks
from hotsos.core.plugins.openvswitch.ovn import OVNBase
from hotsos.core.plugins.openvswitch.ovs import OpenvSwitchBase
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class OpenvSwitchSummary(OpenvSwitchChecks):
    """ Implementation of OpenvSwitch summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ovs = OpenvSwitchBase()
        self.ovn = OVNBase()

    @summary_entry('config', get_min_available_entry_index())
    def summary_ovsdb_config(self):
        _config = {}
        if self.ovs.offload_enabled:
            _config['offload'] = 'enabled'

        _other_config = self.ovs.ovsdb.Open_vSwitch.other_config
        if _other_config:
            _config['other-config'] = _other_config

        _external_ids = self.ovs.ovsdb.Open_vSwitch.external_ids
        if _external_ids:
            _config['external-ids'] = _external_ids

        db_keys = ['inactivity_probe', 'max_backoff']
        for name, db in {'nbdb': self.ovn.nbdb, 'sbdb': self.ovn.sbdb}.items():
            if db:
                _config[name] = {}
                for key in db_keys:
                    val = getattr(db.Connection, key)
                    if val:
                        _config[name][key] = val

        return _config or None

    @summary_entry('bridges', get_min_available_entry_index() + 1)
    def summary_bridges(self):
        bridges = {}
        for bridge in self.ovs.bridges:
            # filter patch/phy ports since they are not generally interesting
            ports = []
            for port in bridge.ports:
                if not isinstance(port, NetworkPort):
                    name = port
                else:
                    name = port.name

                if (name.startswith('patch-') or name.startswith('phy-') or
                        name.startswith('int-')):
                    continue

                if not isinstance(port, NetworkPort):
                    # ovs port so will be just a name
                    ports.append(port)
                    continue

                ports.append(port.to_dict())

            # show total count for br-int since it can have a very large number
            if bridge.name == 'br-int' and ports:
                ports = [f"({len(ports)} ports)"]

            bridges[bridge.name] = ports

        return bridges or None

    @summary_entry('tunnels', get_min_available_entry_index() + 2)
    def summary_tunnels(self):
        return self.ovs.tunnels or None

    @summary_entry('ovn', get_min_available_entry_index() + 3)
    def summary_ovn_resources(self):
        info = {}
        if self.ovn.nbdb:
            routers = self.ovn.nbdb.routers
            switches = self.ovn.nbdb.switches
            info['nbdb'] = {'routers': len(routers),
                            'switches': len(switches)}

        if self.ovn.sbdb:
            ports = []
            cr_ports = []
            for c in self.ovn.sbdb.chassis:
                ports.extend(c.ports)
                cr_ports.extend(c.cr_ports)

            info['sbdb'] = {'chassis': len(self.ovn.sbdb.chassis),
                            'ports': len(ports),
                            'router-gateways': len(cr_ports)}

        return info or None



import glob
import os
from functools import cached_property

from propertree.propertree2 import PTreeLogicalGrouping
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    YDefsSection,
    YDefsContext,
    YPropertyOverrideBase,
    YPropertyMappedOverrideBase,
)
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    OpsUtils,
    YRequirementTypeBase,
)


class YConfigAssertionAttrs(YPropertyOverrideBase):
    """ Logic implementation for config assertion attributes. """
    override_keys = ['key', 'value', 'section', 'ops', 'allow-unset']

    def __bool__(self):
        return bool(self.content)

    def __str__(self):
        return str(self.content)

    @property
    def ops(self):
        """
        This is a list so needs its own property.
        """
        return self.content


class YConfigAssertion(OpsUtils, YPropertyMappedOverrideBase):
    """ Assertion property. Implements the config assertion property used to
    define an assertion on config file contents. """
    override_keys = ['assertion']
    override_members = [YConfigAssertionAttrs]

    @cached_property
    def attrs(self):
        _attrs = {'key': str(self.key), 'allow_unset': False, 'value': None,  # noqa, pylint: disable=E1101
                  'section': None}
        if self.value is not None:  # pylint: disable=E1101
            if isinstance(self.value, bool):  # pylint: disable=E1101
                value = self.value  # pylint: disable=E1101
            else:
                try:
                    value = int(self.value)  # pylint: disable=E1101
                except TypeError:
                    value = str(self.value)  # pylint: disable=E1101

            _attrs['value'] = value

        if self.allow_unset is not None:  # pylint: disable=E1101
            _attrs['allow_unset'] = bool(self.allow_unset)  # noqa, pylint: disable=E1101

        if self.section is not None:  # pylint: disable=E1101
            _attrs['section'] = str(self.section)  # pylint: disable=E1101

        if self.ops:  # pylint: disable=E1101
            _attrs['ops'] = self.ops.ops  # pylint: disable=E1101
        else:
            _attrs['ops'] = [['eq', _attrs['value']]]

        return _attrs

    @property
    def result(self):
        handlers = self.context.assertions_ctxt['cfg_handlers']
        if not handlers:
            log.debug("no handlers found, assuming paths do not exist - "
                      "returning False")
            return False

        key = self.attrs['key']
        value = self.attrs['value']
        section = self.attrs['section']
        ops = self.attrs['ops']
        allow_unset = self.attrs['allow_unset']

        log.debug("%s: key=%s, value=%s, section=%s, ops=%s",
                  self.__class__.__name__, key, value, section, ops)
        for n, cfg_obj in enumerate(handlers):
            if section:
                actual = cfg_obj.get(key, section=section)
            else:
                actual = cfg_obj.get(key)

            _result = allow_unset
            if actual is not None:
                _result = self.apply_ops(ops, opinput=actual,
                                         normalise_value_types=True)
            log.debug("assertion: %s (%s) %s result=%s",
                      key, actual, self.ops_to_str(ops or []),
                      _result)
            if not _result and n < len(handlers) - 1:
                log.debug("assertion is false and there are more configs to "
                          "check so trying next")
                continue

            cache = self.context.assertions_ctxt['cache']
            msg = f"{key} {self.ops_to_str(ops or [])}/actual=\"{actual}\""
            if cache.assertion_results is not None:
                cache.set('assertion_results',
                          f"{cache.assertion_results}, {msg}")
            else:
                cache.set('assertion_results', msg)

            # NOTE: This can be useful for single assertion checks but should
            # be used with caution since it will only ever store the last
            # config checked.
            cache.set('key', key)
            cache.set('ops', self.ops_to_str(ops or []))
            cache.set('value_actual', actual)

        return _result


class AssertionsLogicalGrouping(PTreeLogicalGrouping):
    """ Logical grouping implementation for assertions. """
    override_autoregister = False

    @classmethod
    def and_stop_on_first_false(cls):
        """
        Override the default behaviour of PTreeLogicalGrouping AND groups.

        We want to always execute all member of all logical op groups so that
        we can have their results in the cached assertion_results.
        """
        return False

    def get_items(self):
        log.debug("%s: fetching assertion items", self.__class__.__name__)
        return super().get_items()

    @staticmethod
    def fetch_item_result(item):
        result = item.result
        log.debug("%s: %s result=%s", AssertionsLogicalGrouping.__name__,
                  item.__class__.__name__, result)
        return result


class YConfigAssertions(YPropertyMappedOverrideBase):
    """ Assertions property. Used to define one more assertion. """
    override_keys = ['assertions']
    override_members = [YConfigAssertion]
    override_logical_grouping_type = AssertionsLogicalGrouping

    @property
    def passes(self):
        log.debug("running assertion set")
        try:
            results = []
            stop_executon = False
            for member in self.members:
                for item in member:
                    result = item.result
                    results.append(result)
                    if AssertionsLogicalGrouping.is_exit_condition_met('and',
                                                                       result):
                        stop_executon = True
                        break

                if stop_executon:
                    break

            result = all(results)
        except Exception:
            log.error("exception caught while running assert set:")
            raise

        return result


class YRequirementTypeConfig(YRequirementTypeBase):
    """
    Config requires type property. Provides support for defining a config type
    requirement and perform checks on configuration.
    """
    override_keys = ['config']
    override_autoregister = True

    @property
    def handler(self):
        return self.get_cls(self.context, self.content['handler'])

    @property
    def path(self):
        return self.content.get('path')

    @property
    def cfg_handlers(self):
        """
        A config assertion must have a handler and optionally one or more path
        defined.

        If the path(s) are provided and do not exist the assertion returns
        False.

        @return: list of config handlers
        """
        paths = self.path
        if not paths:
            log.debug("no config path provided, creating handler %s "
                      "without args", self.handler)
            return [self.handler()]  # pylint: disable=E1102

        if isinstance(paths, str):
            paths = [paths]

        handlers = []
        for path in paths:
            path = os.path.join(HotSOSConfig.data_root, path)
            if os.path.isfile(path):
                handlers.append(self.handler(path))  # pylint: disable=E1102
                continue

            if os.path.isdir(path):
                path = path + '/*'

            for _path in glob.glob(path):
                if os.path.isfile(_path):
                    handlers.append(self.handler(_path))  # noqa, pylint: disable=E1102

        return handlers

    @property
    def assertions(self):
        _assertions = []
        ctxt = YDefsContext({'vars': self.context.vars})
        # make this available to all assertions
        ctxt.assertions_ctxt = {'cfg_handlers': self.cfg_handlers,
                                'cache': self.cache}
        section = YDefsSection('config_assertions',
                               {'assertions': self.content.get('assertions')},
                               override_handlers=self.root.override_handlers,
                               resolve_path=self.override_path,
                               context=ctxt)
        for leaf in section.leaf_sections:
            _assertions.append(leaf.assertions.passes)

        return _assertions

    @property
    @intercept_exception
    def _result(self):
        return all(self.assertions)



import abc

from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
)


class ServiceManagerTypeBase(YRequirementTypeBase):
    """
    Base class for implementations for service manager requirement types such
    as systemd, pebble etc.
    """
    default_op = 'eq'

    @property
    @abc.abstractmethod
    def items_to_check(self):
        """ Return check items we want to iterate over until we find an error.

        The implementing class should have an accompanying implementation of
        ServiceCheckItemsBase and return that as an object here.
        """

    @property
    @intercept_exception
    def _result(self):
        cache_info = {}
        _result = True

        items = self.items_to_check
        if not items.not_installed:
            for svc, settings in items:
                svc_obj = items.installed[svc]
                cache_info[svc] = {'actual': svc_obj.state}
                if settings is None:
                    continue

                _result = self._check_item_settings(svc_obj, settings,
                                                    cache_info, items)
                if not _result:
                    # bail on first fail
                    break
        else:
            log.debug("one or more services not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            # bail on first fail i.e. if any not installed
            _result = False

        # this will represent what we have actually checked
        self.cache.set('services', ', '.join(cache_info))
        svcs = [f"{svc}={settings}" for svc, settings in items]
        log.debug('requirement check: %s (result=%s)',
                  ', '.join(svcs), _result)
        return _result
